from collections.abc import AsyncIterator, Callable
from contextlib import asynccontextmanager, redirect_stdout
from copy import deepcopy
from crystal import server
from crystal.model import Project, Resource, ResourceRevision, RootResource
from crystal.server import _DEFAULT_SERVER_PORT, get_request_url
from crystal.tests.util.controls import click_button, TreeItem
from crystal.tests.util.runner import bg_sleep
from crystal.tests.util.server import (
    assert_does_open_webbrowser_to, extracted_project, fetch_archive_url,
    served_project, WebPage,
)
from crystal.tests.util.skip import skipTest
from crystal.tests.util.tasks import wait_for_download_to_start_and_finish
from crystal.tests.util.wait import DEFAULT_WAIT_PERIOD
from crystal.tests.util.windows import (
    MainWindow, NewRootUrlDialog, OpenOrCreateDialog,
)
from crystal.util.ports import is_port_in_use
from io import StringIO

# TODO: Many serving behaviors are tested indirectly by larger tests
#       in test_workflows.py. Write stubs for all such behaviors
#       and link them to the covering test.

# ------------------------------------------------------------------------------
# Test: Start Server

async def test_given_default_serving_port_in_use_when_start_serving_project_then_finds_alternate_port() -> None:
    if is_port_in_use(_DEFAULT_SERVER_PORT):
        skipTest('_DEFAULT_SERVER_PORT is already in use outside of tests')
    if is_port_in_use(_DEFAULT_SERVER_PORT + 1):
        skipTest('_DEFAULT_SERVER_PORT + 1 is already in use outside of tests')
    
    assert not is_port_in_use(_DEFAULT_SERVER_PORT)
    with served_project('testdata_xkcd.crystalproj.zip', port=_DEFAULT_SERVER_PORT) as sp:
        assert is_port_in_use(_DEFAULT_SERVER_PORT)
        
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            # Create a URL
            if True:
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                assert root_ti.GetFirstChild() is None  # no entities
                
                click_button(mw.new_root_url_button)
                nud = await NewRootUrlDialog.wait_for()
                
                nud.name_field.Value = 'Home'
                nud.url_field.Value = home_url
                nud.do_not_download_immediately()
                await nud.ok()
                (home_ti,) = root_ti.Children
            
            # Download the URL
            home_ti.SelectItem()
            await mw.click_download_button()
            await wait_for_download_to_start_and_finish(mw.task_tree)
            
            # Try to start second server, also on _DEFAULT_SERVER_PORT.
            # Expect it to actually start on (_DEFAULT_SERVER_PORT + 1).
            expected_port = _DEFAULT_SERVER_PORT + 1
            home_ti.SelectItem()
            try:
                with assert_does_open_webbrowser_to(get_request_url(home_url, expected_port, project_default_url_prefix=project.default_url_prefix)):
                    click_button(mw.view_button)
            finally:
                assert is_port_in_use(expected_port)
            
            # Ensure can fetch the revision through the server
            server_page = await fetch_archive_url(home_url, expected_port)
            assert 200 == server_page.status


# ------------------------------------------------------------------------------
# Test: Header Inclusion & Exclusion

async def test_when_serve_page_then_safe_headers_included() -> None:
    SAFE_HEADER_NAME = 'Date'
    assert SAFE_HEADER_NAME.lower() in server._HEADER_ALLOWLIST
    
    async with _xkcd_home_page_served() as (revision, server_page, _):
        # Ensure test data has header
        saved_header_value = revision._get_first_value_of_http_header(SAFE_HEADER_NAME)
        assert saved_header_value is not None
        
        # Ensure header has expected value in served page
        served_header_value = server_page.headers[SAFE_HEADER_NAME]
        assert served_header_value is not None
        assert saved_header_value == served_header_value


async def test_when_serve_page_then_unsafe_headers_excluded() -> None:
    UNSAFE_HEADER_NAME = 'Connection'
    assert UNSAFE_HEADER_NAME.lower() in server._HEADER_DENYLIST
    
    async with _xkcd_home_page_served() as (revision, server_page, _):
        # Ensure test data has header
        saved_header_value = revision._get_first_value_of_http_header(UNSAFE_HEADER_NAME)
        assert saved_header_value is not None
        
        # Ensure header has expected value in served page
        served_header_value = server_page.headers[UNSAFE_HEADER_NAME]
        assert served_header_value is None, (
            f'Header {UNSAFE_HEADER_NAME!r} has '
            f'unexpected value {served_header_value!r}'
        )


async def test_when_serve_page_with_unknown_non_x_header_then_excludes_header_and_prints_warning() -> None:
    UNKNOWN_NON_X_HEADER_NAME = 'Crystal-Test-Header'
    UNKNOWN_NON_X_HEADER_VALUE = 'some_value'
    assert (
        UNKNOWN_NON_X_HEADER_NAME.lower() not in server._HEADER_ALLOWLIST and
        UNKNOWN_NON_X_HEADER_NAME.lower() not in server._HEADER_DENYLIST
    )
    
    # Insert header into test data
    def alter_revision(revision: ResourceRevision) -> None:
        saved_header_value = revision._get_first_value_of_http_header(UNKNOWN_NON_X_HEADER_NAME)
        assert saved_header_value is None
        
        new_metadata = deepcopy(revision.metadata)
        assert new_metadata is not None
        new_metadata['headers'].append(
            [UNKNOWN_NON_X_HEADER_NAME, UNKNOWN_NON_X_HEADER_VALUE])
        revision._alter_metadata(new_metadata)
    
    async with _xkcd_home_page_served(alter_revision) as (revision, server_page, captured_stdout):
        # Ensure header has expected value in served page
        served_header_value = server_page.headers[UNKNOWN_NON_X_HEADER_NAME]
        assert served_header_value is None
        
        # Ensure warning printed
        assert (
            f'*** Ignoring unknown header in archive: {UNKNOWN_NON_X_HEADER_NAME}: {UNKNOWN_NON_X_HEADER_VALUE}'
            in captured_stdout
        )


async def test_when_serve_page_with_unknown_x_header_then_excludes_header_silently() -> None:
    UNKNOWN_X_HEADER_NAME = 'X-Timer'
    assert (
        UNKNOWN_X_HEADER_NAME.lower() not in server._HEADER_ALLOWLIST and
        UNKNOWN_X_HEADER_NAME.lower() not in server._HEADER_DENYLIST
    )
    assert True == server._IGNORE_UNKNOWN_X_HEADERS
    
    async with _xkcd_home_page_served() as (revision, server_page, captured_stdout):
        # Ensure test data has header
        saved_header_value = revision._get_first_value_of_http_header(UNKNOWN_X_HEADER_NAME)
        assert saved_header_value is not None
        
        # Ensure header has expected value in served page
        served_header_value = server_page.headers[UNKNOWN_X_HEADER_NAME]
        assert served_header_value is None
        
        # Ensure warning NOT printed
        assert (
            f'*** Ignoring unknown header in archive: {UNKNOWN_X_HEADER_NAME}: '
            not in captured_stdout
        )


# ------------------------------------------------------------------------------
# Utility

@asynccontextmanager
async def _xkcd_home_page_served(
        alter_revision_func: Callable[[ResourceRevision], None] | None=None,
        ) -> AsyncIterator[tuple[ResourceRevision, WebPage, str]]:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        # Define URLs
        home_url = 'https://xkcd.com/'
        
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
            r = Resource(project, home_url)
            
            revision = r.default_revision()
            assert revision is not None
            
            # Alter revision before fetching it through the server, if applicable
            if alter_revision_func is not None:
                alter_revision_func(revision)
            
            (server_page, captured_stdout_str) = await serve_and_fetch_xkcd_home_page(mw)
            
            yield (revision, server_page, captured_stdout_str)


async def serve_and_fetch_xkcd_home_page(mw: MainWindow) -> tuple[WebPage, str]:
    home_url = 'https://xkcd.com/'
    
    with redirect_stdout(StringIO()) as captured_stdout:
        # Start server
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        home_ti = root_ti.GetFirstChild()
        assert home_ti is not None
        assert f'{home_url} - Home' == home_ti.Text
        home_ti.SelectItem()
        with assert_does_open_webbrowser_to(get_request_url(home_url)):
            click_button(mw.view_button)
        
        # Fetch the revision through the server
        server_page = await fetch_archive_url(home_url)
    
    return (server_page, captured_stdout.getvalue())


# ------------------------------------------------------------------------------



from crystal.util.xos import is_linux, is_windows
import wx


def wrap_static_box_sizer_child(child: wx.Sizer) -> wx.Sizer:
    if is_linux():
        # Add padding around contents of wx.StaticBoxSizer because
        # wxGTK does not do this automatically, unlike macOS and Windows
        container = wx.BoxSizer(wx.VERTICAL)
        container.Add(child, proportion=1, flag=wx.ALL|wx.EXPAND, border=8)
        return container
    elif is_windows():
        # Default padding in Windows is cramped. So add some more.
        container = wx.BoxSizer(wx.VERTICAL)
        container.Add(child, proportion=1, flag=wx.ALL|wx.EXPAND, border=4)
        return container
    else:
        return child


from collections.abc import Callable, Iterator
from contextlib import contextmanager
from crystal.tests import (
    test_bulkheads, test_disk_io_errors, test_do_not_download_groups,
    test_download, test_download_body, test_edit_group, test_edit_root_url,
    test_entitytree, test_file_extension_visibility, test_hibernate,
    test_icons, test_install_to_desktop, test_load_urls, test_log_drawer,
    test_menus, test_new_group, test_new_root_url, test_open_project,
    test_parse_html, test_profile, test_project_migrate, test_readonly_mode,
    test_server, test_shell, test_ssd, test_tasks, test_tasktree,
    test_workflows, test_xthreading,
)
from crystal.tests.util.downloads import delay_between_downloads_minimized
from crystal.tests.util.runner import run_test
from crystal.tests.util.subtests import SubtestFailed
from crystal.util.test_mode import tests_are_running
from crystal.util.xcollections.dedup import dedup_list
from crystal.util.xthreading import bg_affinity
from crystal.util.xtime import sleep_profiled
from crystal.util.xtraceback import _CRYSTAL_PACKAGE_PARENT_DIRPATH
import gc
import os
import sys
import time
import traceback
from typing import Dict, Optional
from unittest import SkipTest
import warnings


def _test_functions_in_module(mod) -> list[Callable]:
    return [
        f for f in mod.__dict__.values() 
        if (
            callable(f) and 
            getattr(f, '__name__', '').startswith('test_') and
            # NOTE: Need to check stringness explicitly to exclude "call" from 
            #       "from unittest.mock import call"
            isinstance(getattr(f, '__name__', ''), str)
        )
    ]

# TODO: Avoid the need to manually enumerate all test modules individually
_TEST_FUNCS = (
    _test_functions_in_module(test_bulkheads) +
    _test_functions_in_module(test_disk_io_errors) +
    _test_functions_in_module(test_do_not_download_groups) +
    _test_functions_in_module(test_download) +
    _test_functions_in_module(test_download_body) +
    _test_functions_in_module(test_edit_group) +
    _test_functions_in_module(test_edit_root_url) +
    _test_functions_in_module(test_entitytree) +
    _test_functions_in_module(test_file_extension_visibility) +
    _test_functions_in_module(test_hibernate) +
    _test_functions_in_module(test_icons) +
    _test_functions_in_module(test_install_to_desktop) +
    _test_functions_in_module(test_load_urls) +
    _test_functions_in_module(test_log_drawer) +
    _test_functions_in_module(test_menus) +
    _test_functions_in_module(test_new_group) +
    _test_functions_in_module(test_new_root_url) +
    _test_functions_in_module(test_open_project) +
    _test_functions_in_module(test_parse_html) +
    _test_functions_in_module(test_profile) +
    _test_functions_in_module(test_project_migrate) +
    _test_functions_in_module(test_readonly_mode) +
    _test_functions_in_module(test_server) +
    _test_functions_in_module(test_shell) +
    _test_functions_in_module(test_ssd) +
    _test_functions_in_module(test_tasks) +
    _test_functions_in_module(test_tasktree) +
    _test_functions_in_module(test_workflows) +
    _test_functions_in_module(test_xthreading) +
    []
)


_TestFuncId = tuple[str, str]  # (module, func_name)


# TODO: Rename "test_names" to something more appropriate,
#       now that items can also refer to test modules (and not just test functions)
@bg_affinity
def run_tests(test_names: list[str]) -> bool:
    """
    Runs automated UI tests, printing a summary report,
    and returning whether the run was OK.
    
    The format of the summary report is designed to be similar
    to that used by Python's unittest module.
    """
    with delay_between_downloads_minimized(), sleep_profiled():
        return _run_tests(test_names)


def _run_tests(test_names: list[str]) -> bool:
    # Ensure ancestor caller did already call set_tests_are_running()
    assert tests_are_running()
    
    result_for_test_func_id = {}  # type: Dict[_TestFuncId, Optional[Exception]]
    start_time = time.time()  # capture
    run_count = 0
    with warnings.catch_warnings(record=True) as warning_list, _warnings_sent_to_ci():
        assert warning_list is not None
        
        for test_func in _TEST_FUNCS:
            if not callable(test_func):
                raise ValueError(f'Test function is not callable: {test_func}')
            test_func_id = (test_func.__module__, test_func.__name__)  # type: _TestFuncId
            test_name = f'{test_func_id[0]}.{test_func_id[1]}'
            
            # Only run test if it was requested (or if all tests are to be run)
            if len(test_names) > 0:
                if test_name not in test_names and test_func.__module__ not in test_names:
                    continue
            run_count += 1
            
            os.environ['CRYSTAL_SCREENSHOT_ID'] = test_name
            
            print('=' * 70)
            print(f'RUNNING: {test_func_id[1]} ({test_func_id[0]}.{test_func_id[1]})')
            print('-' * 70)
            try:
                try:
                    run_test(test_func)
                finally:
                    # Flush any stderr output immediately,
                    # to keep aligned with stdout output from the test
                    sys.stderr.flush()
            except AssertionError as e:
                result_for_test_func_id[test_func_id] = e
                
                traceback.print_exc(file=sys.stdout)
                print('FAILURE')
            except SkipTest as e:
                result_for_test_func_id[test_func_id] = e
                
                print(f'SKIP ({str(e)})')
            except Exception as e:
                result_for_test_func_id[test_func_id] = e
                
                if not isinstance(e, SubtestFailed):
                    traceback.print_exc(file=sys.stdout)
                print(f'ERROR ({e.__class__.__name__})')
            else:
                result_for_test_func_id[test_func_id] = None
                
                print('OK')
            print()
            
            # Garbage collect, running any finalizers in __del__() early,
            # such as the warnings printed by ListenableMixin
            gc.collect()
    
    end_time = time.time()  # capture
    delta_time = end_time - start_time
    
    failure_count = 0
    error_count = 0
    skip_count = 0
    failed_test_names = []
    for (test_func_id, result) in result_for_test_func_id.items():
        if result is None:
            pass
        elif isinstance(result, SkipTest):
            skip_count += 1
        else:
            if isinstance(result, AssertionError):
                failure_count += 1
            else:
                error_count += 1
            
            test_name = f'{test_func_id[0]}.{test_func_id[1]}'
            failed_test_names.append(test_name)
    
    is_ok = (failure_count + error_count) == 0
    
    suffix_parts = []
    if failure_count != 0:
        suffix_parts.append(f'failures={failure_count}')
    if error_count != 0:
        suffix_parts.append(f'errors={error_count}')
    if skip_count != 0:
        suffix_parts.append(f'skipped={skip_count}')
    suffix = (
        f' ({", ".join(suffix_parts)})'
        if len(suffix_parts) != 0
        else ''
    )
    
    print('=' * 70)
    print('SUMMARY')
    print('-' * 70)
    for result in result_for_test_func_id.values():
        if result is None:
            print('.', end='')
        elif isinstance(result, AssertionError):
            print('F', end='')
        elif isinstance(result, SkipTest):
            if str(result).startswith('covered by:'):
                print('c', end='')
            else:
                print('s', end='')
        else:
            print('E', end='')
    print()
    
    print('-' * 70)
    print(f'Ran {run_count} tests in {"%.3f" % delta_time}s')
    print()
    print(f'{"OK" if is_ok else "FAILURE"}{suffix}')
    
    # Print warnings, if any
    if len(warning_list) >= 1:
        print()
        print('Warnings:')
        warning_strs = []
        for w in warning_list:
            if w.filename.startswith(_CRYSTAL_PACKAGE_PARENT_DIRPATH):
                short_filepath = os.path.relpath(w.filename, start=_CRYSTAL_PACKAGE_PARENT_DIRPATH)
            else:
                short_filepath = w.filename
            w_str = warnings.formatwarning(w.message, w.category, short_filepath, w.lineno, w.line)
            warning_strs.append(w_str)
        for w_str in sorted(dedup_list(warning_strs)):
            print('- ' + w_str, end='')
    
    # Print command to rerun failed tests
    if len(failed_test_names) != 0:
        print()
        print('Rerun failed tests with:')
        print(f'$ crystal --test {" ".join(failed_test_names)}')
        print()
    
    # Play bell sound in terminal
    print('\a', end='', flush=True)
    
    return is_ok


@contextmanager
def _warnings_sent_to_ci() -> Iterator[None]:
    if not _running_in_ci():
        yield
        return
    
    super_showwarning = warnings.showwarning  # capture
    
    def showwarning(message, category, filename, lineno, file=None, line=None):
        # Try to reformat `filename` to use the Linux format so that warning
        # annotations are associated with the correct file
        # 
        # Depending on OS, `filename` initially looks like:
        #     - macOS: setup/dist/Crystal.app/Contents/Resources/lib/python38.zip/crystal/tests/index.py
        #     - Linux: src/crystal/tests/index.py
        #     - Windows: crystal\tests\index.pyc
        filename_parts = filename.split(os.path.sep)
        if 'crystal' in filename_parts:
            filename_parts = ['src'] + filename_parts[filename_parts.index('crystal'):]
            if filename_parts[-1].endswith('.pyc'):
                filename_parts[-1] = filename_parts[-1][:-1]  # convert .pyc ending to .py
            filename = '/'.join(filename_parts)  # reinterpret
        
        # Create warning annotation in GitHub Action's [Summary > Annotations] section
        # 
        # Syntax: https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#example-setting-a-warning-message
        print(f'::warning file={filename},line={lineno}::{message}')
        
        return super_showwarning(message, category, filename, lineno, file, line)
    
    warnings.showwarning = showwarning
    try:
        yield
    finally:
        warnings.showwarning = super_showwarning


def _running_in_ci() -> bool:
    return os.environ.get('GITHUB_ACTIONS') == 'true'



import socket

_LOCALHOST = '127.0.0.1'


def is_port_in_use(port: int, hostname: str=_LOCALHOST) -> bool:
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        try:
            result = s.connect_ex((hostname, port))
        except:
            raise ValueError(f'socket.connect_ex({(hostname, port)}) failed')
        if result == 0:
            return True
        else:
            return False
    finally:
        s.close()


def is_port_in_use_error(e: Exception) -> bool:
    # macOS: [Errno 48] Address already in use
    # Linux: [Errno 98] Address already in use
    if isinstance(e, OSError) and 'Address already in use' in str(e):
        return True
    return False



from collections.abc import Iterator
from contextlib import contextmanager
from crystal.browser.icons import TREE_NODE_ICONS
from crystal.task import (
    CrashedTask, DownloadResourceGroupMembersTask, RootTask,
    SCHEDULING_STYLE_SEQUENTIAL, Task,
)
from crystal.ui.tree2 import NodeView, NULL_NODE_VIEW, TreeView
from crystal.ui.tree import NodeView as NodeView1
from crystal.util.bulkheads import (
    capture_crashes_to_bulkhead_arg as capture_crashes_to_task_arg,
)
from crystal.util.bulkheads import Bulkhead, BulkheadCell, capture_crashes_to
from crystal.util.bulkheads import capture_crashes_to_bulkhead_arg
from crystal.util.bulkheads import capture_crashes_to_stderr, CrashReason
from crystal.util.wx_bind import bind
from crystal.util.wx_treeitem_gettooltip import (
    EVT_TREE_ITEM_GETTOOLTIP, GetTooltipEvent,
)
from crystal.util.xcollections.lazy import (
    AppendableLazySequence, UnmaterializedItemError,
)
from crystal.util.xthreading import (
    fg_call_and_wait, fg_call_later, is_foreground_thread,
)
from crystal.util.xtraceback import format_exception_for_user
from typing import List, Optional, Tuple
import wx

_ID_DISMISS = 101


class TaskTree:
    """
    View controller for the task tree
    """
    def __init__(self, parent_peer: wx.Window, root_task: Task) -> None:
        self.tree = TreeView(parent_peer, name='cr-task-tree')
        self.tree.delegate = self
        
        self.root = TaskTreeNode(root_task)
        self.tree.root = self.root.tree_node
        
        self._right_clicked_node = None  # type: Optional[TaskTreeNode]
        
        self.tree.peer.SetInitialSize((750, 200))
        
        bind(self.peer, wx.EVT_MOTION, self._on_mouse_motion)
        # For tests only
        bind(self.peer, EVT_TREE_ITEM_GETTOOLTIP, self._on_get_tooltip_event)
    
    @property
    def peer(self) -> wx.TreeCtrl:
        """The wx.TreeCtrl controlled by this class."""
        return self.tree.peer
    
    # === Event: Right Click ===
    
    @capture_crashes_to_stderr
    def on_right_click(self, event: wx.MouseEvent, node_view: NodeView) -> None:
        node = TaskTreeNode.for_node_view(node_view)
        self._right_clicked_node = node
        
        # Create popup menu
        menu = wx.Menu()
        bind(menu, wx.EVT_MENU, self._on_popup_menuitem_selected)
        if node is not None and isinstance(node.task, CrashedTask):
            menu.Append(_ID_DISMISS, node.task.dismiss_action_title)
        elif node is not None and self._is_dismissable_top_level_task(node.task):
            menu.Append(_ID_DISMISS, 'Dismiss')
        else:
            menu.Append(_ID_DISMISS, 'Dismiss')
            menu.Enable(_ID_DISMISS, False)
        
        # Show popup menu
        self.peer.PopupMenu(menu)
        menu.Destroy()
    
    def _on_popup_menuitem_selected(self, event: wx.MenuEvent) -> None:
        node = self._right_clicked_node
        assert node is not None
        
        item_id = event.GetId()
        if item_id == _ID_DISMISS:
            if isinstance(node.task, CrashedTask):
                node.task.dismiss()
            elif self._is_dismissable_top_level_task(node.task):
                self._dismiss_top_level_task(node.task)
            else:
                raise AssertionError(f'Do not know how to dismiss: {node.task}')
    
    @classmethod
    def _is_dismissable_top_level_task(cls, task: Task) -> bool:
        return (
            task.crash_reason is not None and
            isinstance(task.parent, RootTask) and
            task.parent.crash_reason is None
        )
    
    @classmethod
    def _dismiss_top_level_task(cls, task: Task) -> None:
        if not cls._is_dismissable_top_level_task(task):
            raise ValueError()
        if not task.complete:
            task.finish()
    
    # === Event: Mouse Motion, Get Tooltip ===
    
    # Update the tooltip whenever hovering the mouse over a tree node icon
    @capture_crashes_to_stderr
    def _on_mouse_motion(self, event: wx.MouseEvent) -> None:
        (tree_item_id, hit_flags) = self.peer.HitTest(event.Position)
        TREEITEM_BODY_FLAGS = (
            wx.TREE_HITTEST_ONITEMICON |
            wx.TREE_HITTEST_ONITEMLABEL
        )
        if (hit_flags & TREEITEM_BODY_FLAGS) != 0:
            new_tooltip = self._tooltip_for_tree_item_id(tree_item_id)
        else:
            new_tooltip = None
        
        if new_tooltip is None:
            self.peer.UnsetToolTip()
        else:
            self.peer.SetToolTip(new_tooltip)
    
    @capture_crashes_to_stderr
    def _on_get_tooltip_event(self, event: wx.Event) -> None:
        event.tooltip_cell[0] = self._tooltip_for_tree_item_id(event.tree_item_id)
    
    def _tooltip_for_tree_item_id(self, tree_item_id: wx.TreeItemId) -> str | None:
        node_view = self.peer.GetItemData(tree_item_id)  # type: NodeView
        node = TaskTreeNode.for_node_view(node_view)
        if node is None:
            return None
        return node.tooltip
    
    # === Dispose ===
    
    def dispose(self) -> None:
        self.root.dispose()
        self.tree.dispose()


class TaskTreeNode:
    """
    View controller for an individual node of the task tree.
    """
    # The following limits are enforced for SCHEDULING_STYLE_SEQUENTIAL tasks only
    _MAX_LEADING_COMPLETE_CHILDREN = 5
    _MAX_VISIBLE_CHILDREN = 100
    
    _CRASH_TEXT_COLOR = wx.Colour(255, 0, 0)  # red
    _CRASH_SUBTITLE = 'Something went wrong'
    
    # Optimize per-instance memory use, since there may be very many TaskTreeNode objects
    __slots__ = (
        'task',
        'tree_node',
        '_num_visible_children',
        '_visible_children_offset',
        '_first_incomplete_child_index',
        '_ignore_complete_events',
        '_crash_reason_and_tooltip',
    )
    
    def __init__(self, task: Task) -> None:
        self.task = task
        
        # TODO: Export a listener reference to TaskTreeNode later in this function,
        #       after the inner NodeView is created.
        #       It seems risky to export a reference to an incomplete TaskTreeNode.
        if task._use_extra_listener_assertions:
            assert self not in self.task.listeners
        if not task.complete:
            self.task.listeners.append(self)
        
        self.tree_node = NodeView()
        self.tree_node.delegate = self
        if self.task.icon_name is not None:
            self.tree_node.icon_set = (
                (wx.TreeItemIcon_Normal, TREE_NODE_ICONS()[self.task.icon_name]),
            )
        self.tree_node.title = self.task.title
        self.tree_node.subtitle = self._calculate_tree_node_subtitle(
            self.task.subtitle,
            self.task.crash_reason)
        self.tree_node.text_color = self._calculate_tree_node_text_color(self.task.crash_reason)
        self.tree_node.bold = self._calculate_tree_node_bold(self.task.crash_reason)
        self.tree_node.expandable = not callable(task)
        
        self._num_visible_children = 0
        self._visible_children_offset = 0
        self._first_incomplete_child_index = 0
        self._ignore_complete_events = False
        self._crash_reason_and_tooltip = None  # type: Optional[Tuple[BaseException, str]]
        
        # NOTE: Transition to foreground thread here BEFORE making very many
        #       calls to self.task_did_append_child() so that we don't need to
        #       make very many thread transitions in that function
        task_children_count = len(self.task.children)  # capture
        @capture_crashes_to(task)
        def fg_task() -> None:
            # Update current progress dialog, if found,
            # in preparation for appending tasks
            # 
            # HACK: Reaches into a progress dialog managed elsewhere,
            #       in MainWindow._on_download_entity()
            progress_dialog_old_message = None  # type: Optional[str]
            if DownloadResourceGroupMembersTask._LAZY_LOAD_CHILDREN:
                progress_dialog = None  # type: Optional[wx.Window]
            else:
                if task_children_count >= 100:
                    assert is_foreground_thread()
                    progress_dialog = wx.FindWindowByName('cr-starting-download')
                    if progress_dialog is not None:
                        assert isinstance(progress_dialog, wx.ProgressDialog)
                        
                        # Try remove elapsed time from progress dialog,
                        # since we won't be able to keep it up to date soon
                        # 
                        # NOTE: This has no effect on macOS
                        progress_dialog.WindowStyleFlag &= ~wx.PD_ELAPSED_TIME
                        
                        # Change progress dialog message
                        progress_dialog_old_message = progress_dialog.Message
                        progress_dialog.Pulse(f'Adding {task_children_count:n} tasks...')
                else:
                    progress_dialog = None
            
            self.task_did_set_children(self.task, task_children_count)
            
            if progress_dialog is not None:
                assert isinstance(progress_dialog, wx.ProgressDialog)
                
                # Restore old message in progress dialog
                assert is_foreground_thread()
                assert progress_dialog_old_message is not None
                progress_dialog.Pulse(progress_dialog_old_message)
        fg_call_and_wait(fg_task)
    
    @staticmethod
    def for_node_view(node_view: NodeView) -> 'Optional[TaskTreeNode]':
        node = node_view.delegate
        if node is None:
            return None
        elif isinstance(node, TaskTreeNode):
            return node
        else:
            raise AssertionError(
                f'Expected delegate of NodeView to be TaskTreeNode or None but found: {node}')
    
    # === Properties ===
    
    @classmethod
    def _calculate_tree_node_subtitle(cls,
            task_subtitle: str,
            task_crash_reason: CrashReason | None) -> str:
        return (
            task_subtitle
            if task_crash_reason is None
            else cls._CRASH_SUBTITLE
        )
    
    @classmethod
    def _calculate_tree_node_text_color(cls,
            task_crash_reason: CrashReason | None) -> wx.Colour | None:
        return (
            cls._CRASH_TEXT_COLOR
            if task_crash_reason is not None
            else None
        )
    
    @classmethod
    def _calculate_tree_node_bold(cls,
            task_crash_reason: CrashReason | None) -> bool:
        return task_crash_reason is not None
    
    @property
    def tooltip(self) -> str | None:
        """
        The tooltip to display when the mouse hovers over this node.
        """
        if self.task.crash_reason is None:
            # Clear any cached tooltip
            self._crash_reason_and_tooltip = None
            
            return None
        else:
            # Return cached tooltip, if available
            if self._crash_reason_and_tooltip is not None:
                (last_crash_reason, last_tooltip) = self._crash_reason_and_tooltip
                if last_crash_reason is self.task.crash_reason:
                    return last_tooltip
            
            # Calculate and cache tooltip
            tooltip = format_exception_for_user(self.task.crash_reason).rstrip('\n')
            self._crash_reason_and_tooltip = (self.task.crash_reason, tooltip)
            return tooltip
    
    # === Events ===
    
    @capture_crashes_to_task_arg
    def task_subtitle_did_change(self, task: Task) -> None:
        task_subtitle = self.task.subtitle  # capture
        task_crash_reason = self.task.crash_reason  # capture
        @capture_crashes_to(task)
        def fg_task() -> None:
            self.tree_node.subtitle = self._calculate_tree_node_subtitle(task_subtitle, task_crash_reason)
        # NOTE: Use profile=False because no obvious further optimizations exist
        fg_call_later(fg_task, profile=False)
    
    # NOTE: Cannot use @capture_crashes_to_task_arg because would create infinite loop
    @capture_crashes_to_stderr
    def task_crash_reason_did_change(self, task: Task) -> None:
        task_crash_reason = self.task.crash_reason  # capture
        task_subtitle = self.task.subtitle  # capture
        # NOTE: Cannot use @capture_crashes_to(task) because would create infinite loop
        @capture_crashes_to_stderr
        def fg_task() -> None:
            self.tree_node.subtitle = self._calculate_tree_node_subtitle(task_subtitle, task_crash_reason)
            self.tree_node.text_color = self._calculate_tree_node_text_color(task_crash_reason)
            self.tree_node.bold = self._calculate_tree_node_bold(task_crash_reason)
        # NOTE: Use profile=False because no obvious further optimizations exist
        fg_call_later(fg_task, profile=False)
    
    @capture_crashes_to_task_arg
    def task_did_complete(self, task: Task) -> None:
        self.task.listeners.remove(self)
    
    @capture_crashes_to_task_arg
    def task_did_set_children(self, task: Task, child_count: int) -> None:
        if task.scheduling_style == SCHEDULING_STYLE_SEQUENTIAL:
            # Create tree node for each visible task
            visible_child_count = min(child_count, self._MAX_VISIBLE_CHILDREN)
            # NOTE: If `task.children` is an AppendableLazySequence then accessing it can
            #       (1) materialize a child that is already complete, and
            #       (2) call self.task_child_did_complete() on that child
            #           BEFORE the inside of the loop can call
            #           self.task_did_append_child() on this child
            #       So suppress the handling of any such
            #       self.task_child_did_complete() events temporarily.
            with self._complete_events_ignored():
                children_to_append = task.children[:visible_child_count]
                for child in children_to_append:
                    # NOTE: Will also call task_child_did_complete() if the
                    #       child is initially complete, but the event handling
                    #       will be suppressed by the enclosing block
                    self.task_did_append_child(task, child)
            assert self._num_visible_children == visible_child_count
            
            # Create more node as a placeholder for the remaining tasks, if needed
            if visible_child_count < child_count:
                @capture_crashes_to(task)
                def fg_task() -> None:
                    self.tree_node.append_child(_MoreNodeView(child_count - visible_child_count))
                fg_call_later(fg_task)
            
            # Apply deferred child-complete actions
            for child in children_to_append:
                if child.complete:
                    self.task_child_did_complete(task, child)
        else:
            # Greedily create tree node for each task
            for child in task.children:
                # NOTE: Will also call task_child_did_complete() if the
                #       child is initially complete
                self.task_did_append_child(task, child)
    
    @capture_crashes_to_stderr
    def task_did_append_child(self, task: Task, child: Task | None) -> None:
        if isinstance(task, RootTask) and isinstance(child, CrashedTask):
            # Specially, allow RootTask to append a CrashedTask when RootTask is crashed
            bulkhead = BulkheadCell()  # type: Bulkhead
        else:
            bulkhead = task
        self._task_did_append_child(bulkhead, task, child)
    
    @capture_crashes_to_bulkhead_arg
    def _task_did_append_child(self, bulkhead: Bulkhead, task: Task, child: Task | None) -> None:
        if (task.scheduling_style == SCHEDULING_STYLE_SEQUENTIAL and
                self._num_visible_children == self._MAX_VISIBLE_CHILDREN):
            @capture_crashes_to(bulkhead)
            def fg_task() -> None:
                # Find last_more_node, or create if missing
                last_child_tree_node = self.tree_node.children[-1]
                if isinstance(last_child_tree_node, _MoreNodeView):
                    last_more_node = last_child_tree_node
                else:
                    last_more_node = _MoreNodeView()
                    self.tree_node.append_child(last_more_node)
                
                # Increase more_count instead of appending new child
                last_more_node.more_count += 1
            fg_call_later(fg_task)
        else:
            # Lookup (and materialize) child if necessary
            if child is None:
                child = task.children[-1]  # lookup child
            
            # Append tree node for new task child
            child_ttnode = TaskTreeNode(child)
            self._num_visible_children += 1
            @capture_crashes_to(bulkhead)
            def fg_task() -> None:
                self.tree_node.append_child(child_ttnode.tree_node)
            fg_call_later(fg_task)
        
        if child is not None and child.complete:
            self.task_child_did_complete(task, child)
    
    @contextmanager
    def _complete_events_ignored(self) -> Iterator[None]:
        assert self._ignore_complete_events == False
        self._ignore_complete_events = True
        try:
            yield
        finally:
            self._ignore_complete_events = False
    
    @capture_crashes_to_task_arg
    def task_child_did_complete(self, task: Task, child: Task) -> None:
        assert task is self.task
        if task.scheduling_style != SCHEDULING_STYLE_SEQUENTIAL:
            return
        
        if self._ignore_complete_events:
            return
        
        # Locate new index of first incomplete child
        # 
        # NOTE: If `task.children` is an AppendableLazySequence then accessing it can
        #       (1) materialize a child that is already complete, and
        #       (2) call self.task_child_did_complete() on that child
        #       So suppress the handling of any such
        #       self.task_child_did_complete() events temporarily.
        with self._complete_events_ignored():
            while self._first_incomplete_child_index < len(task.children):
                try:
                    child_is_complete = task.children[self._first_incomplete_child_index].complete
                except UnmaterializedItemError:
                    child_is_complete = True
                if not child_is_complete:
                    break
                self._first_incomplete_child_index += 1
        
        # Update desired offset of visible children
        old_visible_children_offset = self._visible_children_offset  # capture, cache
        new_visible_children_offset = max(
            old_visible_children_offset,
            self._first_incomplete_child_index - self._MAX_LEADING_COMPLETE_CHILDREN
        )
        if new_visible_children_offset == old_visible_children_offset:
            return
        self._visible_children_offset = new_visible_children_offset
        
        # Update visible children count
        self._num_visible_children = min(
            len(task.children) - new_visible_children_offset,
            self._MAX_VISIBLE_CHILDREN
        )
        
        # Create new trailing task tree nodes
        trailing_intermediate_nodes_to_add = [
            TaskTreeNode(task.children[trailing_child_index]).tree_node
            for trailing_child_index in range(
                max(
                    old_visible_children_offset + self._MAX_VISIBLE_CHILDREN,
                    new_visible_children_offset
                ),
                min(
                    new_visible_children_offset + self._MAX_VISIBLE_CHILDREN,
                    len(task.children)
                )
            )
        ]
        
        # Prepare: Unmaterialize tasks for old leading task tree nodes
        if isinstance(task.children, AppendableLazySequence):
            lazy_task_children = task.children  # capture
            leading_intermediate_task_indexes_to_remove = [
                leading_child_index
                for leading_child_index in range(
                    old_visible_children_offset,
                    new_visible_children_offset
                )
            ]
            def unmaterialize_tasks() -> None:
                for i in leading_intermediate_task_indexes_to_remove:
                    lazy_task_children.unmaterialize(i)
        else:
            def unmaterialize_tasks() -> None:
                pass
        
        # Update visible children
        task_children_count = len(task.children)  # capture
        @capture_crashes_to(task)
        def fg_task() -> None:
            # Find (first_more_node, intermediate_nodes, last_more_node)
            intermediate_nodes = list(self.tree_node.children)
            if len(intermediate_nodes) == 0:
                # Node is disposed. Abort.
                assert self.tree_node.peer is None
                return
            if isinstance(intermediate_nodes[0], _MoreNodeView):
                first_more_node = intermediate_nodes[0]
                del intermediate_nodes[0]
            else:
                first_more_node = _MoreNodeView()
            if isinstance(intermediate_nodes[-1], _MoreNodeView):
                last_more_node = intermediate_nodes[-1]
                del intermediate_nodes[-1]
            else:
                last_more_node = _MoreNodeView()
            
            # Revise (first_more_node, intermediate_nodes, last_more_node)
            first_more_node.more_count = new_visible_children_offset
            del intermediate_nodes[:(new_visible_children_offset - old_visible_children_offset)]
            intermediate_nodes.extend(trailing_intermediate_nodes_to_add)
            # NOTE: Internally asserts value is >= 0
            last_more_node.more_count = (
                task_children_count - len(intermediate_nodes) - first_more_node.more_count
            )
            
            # Commit changes to the children list
            new_children = []  # type: List[NodeView1]
            if first_more_node.more_count != 0:
                new_children.append(first_more_node)
            new_children.extend(intermediate_nodes)
            if last_more_node.more_count != 0:
                new_children.append(last_more_node)
            self.tree_node.children = new_children  # type: ignore[misc]
            
            # Perform: Unmaterialize tasks for old leading task tree nodes
            unmaterialize_tasks()
        fg_call_later(fg_task)
    
    @capture_crashes_to_task_arg
    def task_did_clear_children(self,
            task: Task,
            child_indexes: list[int] | None=None
            ) -> None:
        assert task is self.task
        if child_indexes is None:
            self._num_visible_children = 0
            @capture_crashes_to(task)
            def fg_task() -> None:
                self.tree_node.children = []  # type: ignore[misc]
            fg_call_later(fg_task)
        else:
            if self.task.scheduling_style == SCHEDULING_STYLE_SEQUENTIAL:
                raise NotImplementedError()
            self._num_visible_children -= len(child_indexes)
            @capture_crashes_to(task)
            def fg_task() -> None:
                self.tree_node.children = [  # type: ignore[misc]
                    c
                    for (i, c) in enumerate(self.tree_node.children)
                    if i not in child_indexes
                ]
            fg_call_later(fg_task)
    
    # === Dispose ===
    
    def dispose(self) -> None:
        self.task.listeners.remove(self)
        self.tree_node.dispose()
        self.tree_node = NULL_NODE_VIEW
    
    # === Utility ===
    
    def __repr__(self) -> str:
        return f'TaskTreeNode({self.task!r})'


class _MoreNodeView(NodeView):
    def __init__(self, more_count: int=0) -> None:
        self._more_count = -1
        
        super().__init__()
        self.icon_set = (
            (wx.TreeItemIcon_Normal, TREE_NODE_ICONS()['entitytree_more']),
        )
        self.more_count = more_count  # sets self.title too
    
    def _get_more_count(self) -> int:
        return self._more_count
    def _set_more_count(self, more_count: int) -> None:
        if more_count == self._more_count:
            return
        if more_count < 0:
            raise ValueError(f'Cannot set negative more_count: {more_count}')
        self._more_count = more_count
        self.title = f'{self._more_count:n} more'
    more_count = property(_get_more_count, _set_more_count)



from collections.abc import Iterator
from contextlib import contextmanager, redirect_stderr
from crystal.tests.util.subtests import SubtestsContext, with_subtests
from crystal.util.profile import (
    ignore_runtime_from_enclosing_warn_if_slow, warn_if_slow,
)
import gc
import io
from unittest.mock import patch


@with_subtests
def test_warn_if_slow_context_does_warn_appropriately(subtests: SubtestsContext) -> None:
    with _gc_disabled():
        with subtests.test(context_execution_time='fast'):
            with patch('time.time', side_effect=[0.0, 0.1]):
                with redirect_stderr(io.StringIO()) as captured_stderr:
                    with warn_if_slow(
                            'Inserting links',
                            max_duration=1.0,
                            message='1 link from https://xkcd.com/',
                            enabled=True):
                        pass
            assert '' == captured_stderr.getvalue(), \
                'Expected no warning to be printed for a fast-executing context'
        
        with subtests.test(context_execution_time='slow'):
            with patch('time.time', side_effect=[0.0, 1.1]):
                with redirect_stderr(io.StringIO()) as captured_stderr:
                    with warn_if_slow(
                            'Inserting links',
                            max_duration=1.0,
                            message='1 link from https://xkcd.com/',
                            enabled=True):
                        pass
            assert (
                '*** Inserting links took 1.10s to execute: 1 link from https://xkcd.com/\n' == 
                captured_stderr.getvalue()
            ), 'Expected a warning to be printed for a slow-executing context'


@with_subtests
def test_warn_if_slow_context_does_exclude_inner_context_runtime(subtests: SubtestsContext) -> None:
    with _gc_disabled():
        with subtests.test(nesting_level=1):
            with subtests.test(
                    inner_context_type='warn_if_slow'):
                with patch('time.time', side_effect=[0.0, 0.1, 1.2, 1.3]):
                    with redirect_stderr(io.StringIO()) as captured_stderr:
                        with warn_if_slow('Outer context', max_duration=1.0, message='Outer'):
                            with warn_if_slow('Inner context', max_duration=1.0, message='Inner'):
                                pass
                assert (
                    '*** Inner context took 1.10s to execute: Inner\n' == 
                    captured_stderr.getvalue()
                ), 'Expected a warning to be printed by inner context but not the outer context'
            
            with subtests.test(
                    inner_context_type='ignore_runtime_from_enclosing_warn_if_slow'):
                with patch('time.time', side_effect=[0.0, 0.1, 1.2, 1.3]):
                    with redirect_stderr(io.StringIO()) as captured_stderr:
                        with warn_if_slow('Outer context', max_duration=1.0, message='Outer'):
                            with ignore_runtime_from_enclosing_warn_if_slow():
                                pass
                assert (
                    '' == 
                    captured_stderr.getvalue()
                ), 'Expected no warning to be printed by outer context'
        
        with subtests.test(nesting_level=2):
            with subtests.test(
                    inner_context_type='warn_if_slow',
                    inner_inner_context_type='warn_if_slow'):
                with patch('time.time', side_effect=[0.0, 0.1, 0.2, 1.3, 1.4, 1.5]):
                    with redirect_stderr(io.StringIO()) as captured_stderr:
                        with warn_if_slow('Outer context', max_duration=1.0, message='Outer'):
                            with warn_if_slow('Inner context', max_duration=1.0, message='Inner'):
                                with warn_if_slow('Inner inner context', max_duration=1.0, message='Inner inner'):
                                    pass
                assert (
                    '*** Inner inner context took 1.10s to execute: Inner inner\n' == 
                    captured_stderr.getvalue()
                ), 'Expected a warning to be printed by inner inner context but not by any other context'
            
            with subtests.test(
                    inner_context_type='warn_if_slow',
                    inner_inner_context_type='ignore_runtime_from_enclosing_warn_if_slow'):
                with patch('time.time', side_effect=[0.0, 0.1, 0.2, 1.3, 1.4, 1.5]):
                    with redirect_stderr(io.StringIO()) as captured_stderr:
                        with warn_if_slow('Outer context', max_duration=1.0, message='Outer'):
                            with warn_if_slow('Inner context', max_duration=1.0, message='Inner'):
                                with ignore_runtime_from_enclosing_warn_if_slow():
                                    pass
                assert (
                    '' == 
                    captured_stderr.getvalue()
                ), 'Expected no warning to be printed by any context'


@contextmanager
def _gc_disabled() -> Iterator[None]:
    # Disable garbage collection because it uses time.time(),
    # and callers are mocking the use of time.time()
    gc.disable()
    try:
        yield
    finally:
        gc.enable()



from collections.abc import Callable, Iterator
from contextlib import contextmanager
from crystal.util.xos import is_mac_os
from functools import wraps
import os
import sys
from typing import TypeVar
from typing_extensions import ParamSpec

_P = ParamSpec('_P')
_T = TypeVar('_T')


def take_error_screenshot() -> None:
    """
    Takes a screenshot, just before an exception that will likely fail
    an automated test is about to be thrown.
    """
    # Take screenshot if screenshots directory path defined
    screenshots_dirpath = os.environ.get('CRYSTAL_SCREENSHOTS_DIRPATH')
    if screenshots_dirpath is None:
        return
    
    # Try import and configure screenshot-related libraries
    try:
        import PIL
    except ImportError:
        print('*** Unable to save screenshot because PIL is not available, which pyscreeze depends on.', file=sys.stderr)
        return
    try:
        import pyscreeze
    except ImportError:
        print('*** Unable to save screenshot because pyscreeze is not available.', file=sys.stderr)
        return
    # HACK: Force pyscreeze to use 'screencapture' tool on macOS
    # in _screenshot_osx() rather than ImageGrab.grab(),
    # which doesn't seem to work on macOS 12+
    if is_mac_os():
        pyscreeze.PIL__version__ = [1, 0, 0]  # pretend PIL is old version
    
    os.makedirs(screenshots_dirpath, exist_ok=True)
    
    screenshot_filename = os.environ.get('CRYSTAL_SCREENSHOT_ID', 'screenshot') + '.png'
    screenshot_filepath = os.path.abspath(os.path.join(screenshots_dirpath, screenshot_filename))
    print('*** Saving screenshot to: ' + screenshot_filepath, file=sys.stderr)
    pyscreeze.screenshot(screenshot_filepath)
    if not os.path.exists(screenshot_filepath):
        print('*** Screenshot not saved. Is this macOS and the Screen Recording permission has not been granted?', file=sys.stderr)


@contextmanager
def screenshot_if_raises() -> Iterator[None]:
    """
    Context that will take a screenshot if an exception is raised within it.
    """
    try:
        yield
    except:
        take_error_screenshot()
        raise


def screenshot_if_raises_deco(callable: Callable[_P, _T]) -> Callable[_P, _T]:
    """
    Decorates a function to make it take a screenshot whenever it raises an exception.
    """
    @wraps(callable)
    def wrapped(*args: _P.args, **kwargs: _P.kwargs) -> _T:
        with screenshot_if_raises():
            return callable(*args, **kwargs)
    return wrapped



from collections.abc import Callable
from crystal.util.lazymap import lazymap
from sortedcontainers import SortedKeyList, SortedList
from typing import Generic, TypeVar

_E1 = TypeVar('_E1')
_E2 = TypeVar('_E2')


def bisect_key_right(
        a: 'SortedList[_E1]',
        x: _E2,
        order_preserving_key: Callable[[_E1], _E2]
        ) -> int:
    """
    Locate the insertion point for `x` in `a` to maintain sorted order,
    returning an insertion point which comes after (to the right of) any
    existing entries of x in a.
    
    `order_preserving_key` specifies a key function of one argument that 
    is used to extract a comparison key from each element in the array.
    To support searching complex records, the key function is not applied
    to the `x` value.
    
    When applying `order_preserving_key` to every element of `a` in order,
    it must result in a list that is itself sorted. If this condition is
    not true then the bisection result is undefined.
    """
    return _LazySortedKeyListAdapter(a, order_preserving_key).bisect_key_right(x)


class _LazySortedKeyListAdapter(Generic[_E1]):  # behaves like a SortedKeyList[_E1]
    def __init__(self,
            a: 'SortedList[_E1]',
            order_preserving_key: Callable[[_E1], _E2]
            ) -> None:
        self._a = a
        
        # Define internals of SortedKeyList
        self._maxes = lazymap(
            a._maxes,
            key=order_preserving_key
        )
        self._keys = lazymap(
            a._lists,
            key=lambda sublist: lazymap(
                sublist,
                key=order_preserving_key
            )
        )
    
    def __getattr__(self, attr_name: str):
        # Obtain remaining internals of SortedKeyList from the base SortedList
        return getattr(self._a, attr_name)
    
    def bisect_key_right(self, x: _E2) -> int:
        return SortedKeyList.bisect_key_right(self, x)



from collections.abc import Callable
from crystal.util.wx_bind import bind
from crystal.util.wx_dialog import (
    position_dialog_initially, set_dialog_or_frame_icon_if_appropriate,
)
from crystal.util.xos import is_wx_gtk
import wx

_WINDOW_INNER_PADDING = 10


class BetterMessageDialog(wx.Dialog):
    """
    Implements a version of wx.MessageDialog that allows the button names
    to be customized. It also allows an optional checkbox.
    """
    def __init__(self, 
            parent: wx.Window, 
            message: str, 
            title: str, 
            style,
            *, checkbox_label: str | None=None,
            on_checkbox_clicked: Callable[[wx.CommandEvent], None] | None=None,
            yes_label: str | None=None, 
            no_label: str | None=None, 
            escape_is_cancel: bool=False,
            name: str | None=None,
            ) -> None:
        """
        Arguments:
        * parent -- parent window.
        * message -- the message displayed in the dialog.
        * title -- the title displayed in the dialog's titlebar.
        * style -- the set of buttons to display. See wx.Dialog.CreateButtonSizer() for all options.
        * yes_label -- label for the wx.YES button.
        * no_label -- label for the wx.NO button.
        """
        if name is None:
            super().__init__(parent, title=title)
        else:
            super().__init__(parent, title=title, name=name)
        set_dialog_or_frame_icon_if_appropriate(self)
        
        self_sizer = wx.BoxSizer(wx.VERTICAL); self.SetSizer(self_sizer)
        
        bind(self, wx.EVT_BUTTON, self._on_button)
        
        message_label = wx.StaticText(self, label=message)
        message_label.Wrap(400)
        self_sizer.Add(
            message_label,
            flag=wx.ALL,
            border=_WINDOW_INNER_PADDING)
        if checkbox_label is None:
            self._checkbox = None
        else:
            self._checkbox = wx.CheckBox(
                self, label=checkbox_label,
                **(dict(name=f'{name}__checkbox') if name else dict()))
            if on_checkbox_clicked is not None:
                bind(self._checkbox, wx.EVT_CHECKBOX, on_checkbox_clicked)
            self_sizer.Add(
                self._checkbox,
                flag=wx.LEFT | wx.BOTTOM | wx.ALIGN_LEFT,
                border=_WINDOW_INNER_PADDING)
        self_sizer.Add(
            self.CreateButtonSizer(style),
            flag=wx.BOTTOM | wx.ALIGN_RIGHT,
            border=_WINDOW_INNER_PADDING)
        
        # Customize button titles
        if yes_label is not None:
            self.FindWindow(id=wx.ID_YES).SetLabel(yes_label)
        if no_label is not None:
            self.FindWindow(id=wx.ID_NO).SetLabel(no_label)
        
        if escape_is_cancel:
            self.SetEscapeId(wx.ID_CANCEL)
        
        position_dialog_initially(self)
        
        # HACK: wxGTK won't compute size of first dialog shown correctly
        #       unless it is explicitly shown during a Fit()
        if is_wx_gtk():
            self.Show()
        self.Fit()
        if is_wx_gtk():
            self.Hide()
    
    def IsCheckBoxChecked(self) -> bool:
        if self._checkbox is None:
            raise ValueError()
        return self._checkbox.Value
    
    def _on_button(self, event: wx.CommandEvent) -> None:
        self.EndModal(event.GetId())
        self.Hide()



import bs4
from bs4 import BeautifulSoup
from collections.abc import Callable, Iterable, MutableMapping
import lxml.html
from re import Pattern
from typing import Literal, Optional, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from crystal.doc.html import HtmlParserType


def parse_html(
        html_bytes: bytes,
        from_encoding: str | None,
        parser_type: 'HtmlParserType',
        ) -> 'FastSoup':
    """
    Parses an HTML document, returning a FastSoup object that can be
    examined through a BeautifulSoup-compatible API.
    """
    if parser_type == 'lxml':
        parser = lxml.html.HTMLParser(encoding=from_encoding)
        root = lxml.html.document_fromstring(html_bytes, parser=parser)
        return LxmlFastSoup(root)
    elif parser_type == 'html_parser':
        # TODO: Consider supporting 'html.parser' without using the
        #       real BeautifulSoup API to wrap it, which would probably
        #       be faster to parse.
        # NOTE: Although the native BeautifulSoup API also supports the
        #       'html5lib' parser, no version of Crystal has used it so far.
        return BeautifulFastSoup(BeautifulSoup(
            html_bytes, from_encoding=from_encoding, features='html.parser'))
    else:
        raise ValueError(f'Unrecognized value for parser_type: {parser_type}')


Tag = Union[lxml.html.HtmlElement, bs4.Tag]

FindFunc = Callable[['FastSoup'], Iterable[Tag]]


class FastSoup:  # abstract
    """A parsed HTML or XML document, navigable with a BeautifulSoup-compatible API."""
    
    # === Document ===
    
    def find_all(self, 
            tag_name: str | None=None, 
            **attrs: str | Pattern | Literal[True]
            ) -> Iterable[Tag]:
        raise NotImplementedError()
    
    @classmethod
    def find_all_compile(cls, 
            tag_name: str | None=None, 
            **attrs: str | Pattern | Literal[True]
            ) -> FindFunc:
        raise NotImplementedError()
    
    def find(self, pattern: Literal[True]) -> 'Optional[Tag]':
        raise NotImplementedError()
    
    def new_tag(self, tag_name: str) -> 'Tag':
        raise NotImplementedError()
    
    def __str__(self) -> str:
        raise NotImplementedError()
    
    # === Tags ===
    
    def tag_name(self, tag: Tag) -> str:
        raise NotImplementedError()
    
    def tag_attrs(self, tag: Tag) -> MutableMapping[str, str | list[str]]:
        raise NotImplementedError()
    
    def tag_string(self, tag: Tag) -> str | None:
        raise NotImplementedError()
    
    def set_tag_string(self, tag: Tag, string: str | None) -> None:
        raise NotImplementedError()
    
    def tag_insert_before(self, tag: Tag, tag2: Tag) -> None:
        raise NotImplementedError()


class BeautifulFastSoup(FastSoup):
    def __init__(self, base: BeautifulSoup) -> None:
        self._base = base
    
    # === Document ===
    
    def find_all(self, 
            tag_name: str | None=None, 
            **attrs: str | Pattern | Literal[True]
            ) -> Iterable[Tag]:
        return self._base.find_all(tag_name, **attrs)  # type: ignore[arg-type]
    
    # NOTE: BeautifulFastSoup doesn't actually support precompiling find_all() queries
    @classmethod
    def find_all_compile(cls, 
            tag_name: str | None=None, 
            **attrs: str | Pattern | Literal[True]
            ) -> FindFunc:
        def find_func(soup: FastSoup) -> Iterable[Tag]:
            if not isinstance(soup, BeautifulFastSoup):
                raise TypeError()
            return soup.find_all(tag_name, **attrs)
        return find_func
    
    def find(self, pattern: Literal[True]) -> 'Optional[Tag]':
        result = self._base.find(pattern)  # type: ignore[arg-type]
        assert not isinstance(result, bs4.NavigableString)
        return result
    
    def new_tag(self, tag_name: str) -> 'Tag':
        return self._base.new_tag(tag_name)
    
    def __str__(self) -> str:
        return str(self._base)
    
    # === Tags ===
    
    def tag_name(self, tag: Tag) -> str:
        assert isinstance(tag, bs4.Tag)
        return tag.name
    
    def tag_attrs(self, tag: Tag) -> MutableMapping[str, str | list[str]]:
        assert isinstance(tag, bs4.Tag)
        return tag.attrs  # type: ignore[return-value]
    
    def tag_string(self, tag: Tag) -> str | None:
        assert isinstance(tag, bs4.Tag)
        return tag.string
    
    def set_tag_string(self, tag: Tag, string: str | None) -> None:
        assert isinstance(tag, bs4.Tag)
        tag.string = string;
    
    def tag_insert_before(self, tag: Tag, tag2: Tag) -> None:
        assert isinstance(tag, bs4.Tag)
        assert isinstance(tag2, bs4.Tag)
        tag.insert_before(tag2)


class LxmlFastSoup(FastSoup):
    def __init__(self, root: lxml.html.HtmlElement) -> None:
        self._root = root
    
    # === Document ===
    
    def find_all(self, 
            tag_name: str | None=None, 
            **attrs: str | Pattern | Literal[True]
            ) -> Iterable[Tag]:
        return self.find_all_compile(tag_name, **attrs)(self)
    
    @classmethod
    def find_all_compile(cls, 
            tag_name: str | None=None, 
            **attrs: str | Pattern | Literal[True]
            ) -> FindFunc:
        tag_pattern = tag_name if tag_name is not None else '*'
        attr_pattern = ''.join([
            f'[{cls._make_attr_pattern_part(k, v_pat)}]'
            for (k, v_pat) in attrs.items()
            if k != 'string'
        ])
        total_pattern = f'.//{tag_pattern}{attr_pattern}'
        
        re_attrs = [
            (k, v_pat)
            for (k, v_pat) in attrs.items()
            if isinstance(v_pat, Pattern)
        ]
        
        # Filter by tag name, attribute existence, and exact attribute values
        find_func = cls._find_tags_matching_xpath(lxml.etree.XPath(total_pattern))
        # Filter by attribute values matching regular expression patterns
        # NOTE: XPath DOES support regular expressions natively, but I couldn't
        #       get it to accept a pattern that matched both single quotes and
        #       double quotes in the same pattern, even with various kinds of escaping.
        #       Therefore I filter by regular expressions manually.
        for (k, v_pat) in re_attrs:
            if k == 'string':
                find_func = cls._filter_tags_whose_text_matches_pattern(
                    find_func, v_pat)
            else:
                find_func = cls._filter_tags_whose_attrib_matches_pattern(
                    find_func, k, v_pat)
        return find_func
    
    @staticmethod
    def _make_attr_pattern_part(name: str, value_pat: str | Pattern | Literal[True]) -> str:
        if value_pat == True or isinstance(value_pat, Pattern):
            return f'@{name}'
        elif isinstance(value_pat, str):
            return f'@{name}="{value_pat}"'
        else:
            raise ValueError()
    
    @staticmethod
    def _find_tags_matching_xpath(xpath: lxml.etree.XPath) -> FindFunc:
        def find_func(soup: FastSoup) -> Iterable[Tag]:
            if not isinstance(soup, LxmlFastSoup):
                raise TypeError()
            return xpath(soup._root)
        return find_func
    
    @staticmethod
    def _filter_tags_whose_text_matches_pattern(
            find_func: FindFunc,
            v_pat: Pattern
            ) -> FindFunc:
        def new_find_func(soup: FastSoup) -> Iterable[Tag]:
            for tag in find_func(soup):  # type: lxml.html.HtmlElement
                v = tag.text
                if isinstance(v, str) and v_pat.search(v) is not None:
                    yield tag
        return new_find_func
    
    @staticmethod
    def _filter_tags_whose_attrib_matches_pattern(
            find_func: FindFunc,
            k: str,
            v_pat: Pattern
            ) -> FindFunc:
        def new_find_func(soup: FastSoup) -> Iterable[Tag]:
            for tag in find_func(soup):  # type: lxml.html.HtmlElement
                v = tag.attrib[k]
                if isinstance(v, str) and v_pat.search(v) is not None:
                    yield tag
        return new_find_func
    
    def find(self, pattern: Literal[True]) -> 'Optional[Tag]':
        return self._root.find('*')
    
    def new_tag(self, tag_name: str) -> 'Tag':
        return self._root.makeelement(tag_name)
    
    def __str__(self) -> str:
        return lxml.html.tostring(self._root, encoding='unicode')
    
    # === Tags ===
    
    def tag_name(self, tag: Tag) -> str:
        assert isinstance(tag, lxml.html.HtmlElement)
        return tag.tag
    
    def tag_attrs(self, tag: Tag) -> MutableMapping[str, str | list[str]]:
        assert isinstance(tag, lxml.html.HtmlElement)
        return tag.attrib
    
    def tag_string(self, tag: Tag) -> str | None:
        assert isinstance(tag, lxml.html.HtmlElement)
        return tag.text
    
    def set_tag_string(self, tag: Tag, string: str | None) -> None:
        assert isinstance(tag, lxml.html.HtmlElement)
        tag.text = string
    
    def tag_insert_before(self, tag: Tag, tag2: Tag) -> None:
        assert isinstance(tag, lxml.html.HtmlElement)
        assert isinstance(tag2, lxml.html.HtmlElement)
        tag.addprevious(tag2)


def name_of_tag(tag: Tag) -> str:
    if isinstance(tag, lxml.html.HtmlElement):
        return tag.tag
    elif isinstance(tag, bs4.Tag):
        return tag.name
    else:
        raise ValueError()



from crystal.model import Project, Resource, ResourceGroup
from crystal.progress import CancelLoadUrls
# TODO: Consider extracting serve_and_fetch_xkcd_home_page() to utility module
from crystal.tests.test_server import serve_and_fetch_xkcd_home_page
from crystal.tests.util.controls import click_button, TreeItem
from crystal.tests.util.server import extracted_project, served_project
from crystal.tests.util.ssd import database_on_ssd
from crystal.tests.util.subtests import awith_subtests, SubtestsContext
from crystal.tests.util.tasks import wait_for_download_to_start_and_finish
from crystal.tests.util.wait import (
    first_child_of_tree_item_is_not_loading_condition,
    tree_has_no_children_condition, wait_for,
)
from crystal.tests.util.windows import NewGroupDialog, OpenOrCreateDialog
from unittest import skip
from unittest.mock import patch

# === Test: Database not on SSD ===

async def test_given_project_database_not_on_ssd_when_expanding_first_resource_group_node_in_entity_tree_then_loading_urls_progress_dialog_becomes_visible_and_shows_loading_node() -> None:
    with database_on_ssd(False):
        with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath, readonly=True) as (mw, project):
                comic_group = project.get_resource_group('Comics')
                assert comic_group is not None
                
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                
                (comic_group_ti,) = (
                    child for child in root_ti.Children
                    if child.Text.endswith(f'- {comic_group.name}')
                )
                
                # Prepare to press Cancel when LoadUrlsProgressDialog appears
                with patch.object(
                        project._load_urls_progress_listener,
                        'loading_resource',
                        side_effect=CancelLoadUrls) as progress_listener_method:
                    comic_group_ti.Expand()
                    
                    # Ensure shows loading node initially
                    assert True == comic_group_ti.IsExpanded()
                    child = comic_group_ti.GetFirstChild()
                    assert child is not None and child.Text == 'Loading...'
                    
                    # Ensure hides loading node after pressing Cancel
                    await wait_for(lambda: (progress_listener_method.call_count >= 1) or None)
                    await wait_for(lambda: True if not comic_group_ti.IsExpanded() else None)


@skip('covered by: test_given_project_database_not_on_ssd_when_expanding_first_resource_group_node_in_entity_tree_then_loading_urls_progress_dialog_becomes_visible_and_shows_loading_node')
async def test_given_project_database_not_on_ssd_given_expanding_first_resource_group_node_in_entity_tree_and_loading_urls_progress_dialog_is_visible_when_press_cancel_then_hides_dialog_and_hides_loading_node() -> None:
    pass


async def test_given_project_database_not_on_ssd_given_resource_group_node_selected_when_press_download_button_then_loading_urls_progress_dialog_becomes_visible() -> None:
    with database_on_ssd(False):
        with served_project('testdata_xkcd.crystalproj.zip') as sp:
            # Define URLs
            atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
            rss_feed_url = sp.get_request_url('https://xkcd.com/rss.xml')
            feed_pattern = sp.get_request_url('https://xkcd.com/*.xml')
            
            async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
                # Create small resource group (with only 2 members)
                Resource(project, atom_feed_url)
                Resource(project, rss_feed_url)
                feed_group = ResourceGroup(project, 'Feed', feed_pattern)
                
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                
                (feed_group_ti,) = (
                    child for child in root_ti.Children
                    if child.Text.endswith(f'- {feed_group.name}')
                )
                
                # Prepare to press Cancel when LoadUrlsProgressDialog appears
                with patch.object(
                        project._load_urls_progress_listener,
                        'loading_resource',
                        side_effect=CancelLoadUrls) as progress_listener_method:
                    feed_group_ti.SelectItem()
                    click_button(mw.download_button)
                    
                    # Wait for progress dialog to show and for cancel to be pressed
                    await wait_for(lambda: (progress_listener_method.call_count >= 1) or None)
                    
                    # Ensure did not create a download task
                    assert tree_has_no_children_condition(mw.task_tree)() is not None


@skip('covered by: test_given_project_database_not_on_ssd_given_resource_group_node_selected_when_press_download_button_then_loading_urls_progress_dialog_becomes_visible')
async def test_given_project_database_not_on_ssd_given_resource_group_node_selected_and_did_press_download_button_and_loading_urls_progress_dialog_is_visible_when_press_cancel_then_hides_dialog() -> None:
    pass


async def test_given_project_database_not_on_ssd_when_press_new_group_button_then_loading_urls_progress_dialog_becomes_visible() -> None:
    with database_on_ssd(False):
        with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
                # Prepare to press Cancel when LoadUrlsProgressDialog appears
                with patch.object(
                        project._load_urls_progress_listener,
                        'loading_resource',
                        side_effect=CancelLoadUrls) as progress_listener_method:
                    click_button(mw.new_group_button)
                    
                    # Wait for progress dialog to show and for cancel to be pressed
                    await wait_for(lambda: (progress_listener_method.call_count >= 1) or None)
                    
                    # Ensure did not show NewGroupDialog
                    assert NewGroupDialog.window_condition()() is None


@skip('covered by: test_given_project_database_not_on_ssd_when_press_new_group_button_then_loading_urls_progress_dialog_becomes_visible')
async def test_given_project_database_not_on_ssd_given_did_press_new_group_button_and_loading_urls_progress_dialog_is_visible_when_press_cancel_then_hides_dialog_and_add_group_dialog_does_not_appear() -> None:
    pass


# === Test: Database on SSD ===

async def test_given_project_database_on_ssd_when_expanding_any_resource_group_node_in_entity_tree_then_shows_loading_node() -> None:
    with database_on_ssd(True):
        with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath, readonly=True) as (mw, project):
                comic_group = project.get_resource_group('Comics')
                assert comic_group is not None
                
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                
                (comic_group_ti,) = (
                    child for child in root_ti.Children
                    if child.Text.endswith(f'- {comic_group.name}')
                )
                
                # Prepare to spy on whether LoadUrlsProgressDialog appears
                with patch.object(
                        project._load_urls_progress_listener,
                        'loading_resource',
                        wraps=project._load_urls_progress_listener.loading_resource) as progress_listener_method:
                    comic_group_ti.Expand()
                    
                    # Ensure shows loading node initially
                    assert True == comic_group_ti.IsExpanded()
                    child = comic_group_ti.GetFirstChild()
                    assert child is not None and child.Text == 'Loading...'
                    
                    # Wait for expand to complete
                    await wait_for(first_child_of_tree_item_is_not_loading_condition(comic_group_ti))
                    
                    # Ensure did not show LoadUrlsProgressDialog
                    assert 0 == progress_listener_method.call_count


async def test_given_project_database_on_ssd_given_resource_group_node_selected_when_press_download_button_then_download_task_has_subtitle_loading() -> None:
    with database_on_ssd(True):
        with served_project('testdata_xkcd.crystalproj.zip') as sp:
            # Define URLs
            atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
            rss_feed_url = sp.get_request_url('https://xkcd.com/rss.xml')
            feed_pattern = sp.get_request_url('https://xkcd.com/*.xml')
            
            async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
                # Create small resource group (with only 2 members)
                Resource(project, atom_feed_url)
                Resource(project, rss_feed_url)
                feed_group = ResourceGroup(project, 'Feed', feed_pattern)
                
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                
                (feed_group_ti,) = (
                    child for child in root_ti.Children
                    if child.Text.endswith(f'- {feed_group.name}')
                )
                
                # Prepare to spy on whether LoadUrlsProgressDialog appears
                with patch.object(
                        project._load_urls_progress_listener,
                        'loading_resource',
                        wraps=project._load_urls_progress_listener.loading_resource) as progress_listener_method:
                    feed_group_ti.SelectItem()
                    
                    # Wait for download to start and complete
                    await mw.click_download_button()
                    await wait_for_download_to_start_and_finish(mw.task_tree)
                    
                    # Ensure did not show LoadUrlsProgressDialog
                    assert 0 == progress_listener_method.call_count


async def test_given_project_database_on_ssd_when_press_new_group_button_then_add_group_dialog_does_appear() -> None:
    with database_on_ssd(True):
        with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
                # Prepare to spy on whether LoadUrlsProgressDialog appears
                with patch.object(
                        project._load_urls_progress_listener,
                        'loading_resource',
                        wraps=project._load_urls_progress_listener.loading_resource) as progress_listener_method:
                    click_button(mw.new_group_button)
                    
                    # Wait for NewGroupDialog to appear
                    ngd = await NewGroupDialog.wait_for()
                    
                    # Ensure did not show LoadUrlsProgressDialog
                    assert 0 == progress_listener_method.call_count
                    
                    click_button(ngd.cancel_button)


# === Test: No Load Required ===

@awith_subtests
async def test_serve_url_never_requires_loading_urls(subtests: SubtestsContext) -> None:
    for is_ssd in [False, True]:
        with subtests.test(is_ssd=is_ssd), database_on_ssd(is_ssd):
            with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
                # Define URLs
                home_url = 'https://xkcd.com/'
                
                async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath, readonly=True) as (mw, project):
                    root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                    
                    (home_ti,) = (
                        child for child in root_ti.Children
                        if child.Text.endswith(f'- Home')
                    )
                    
                    # Prepare to spy on whether LoadUrlsProgressDialog appears
                    with patch.object(
                            project._load_urls_progress_listener,
                            'loading_resource',
                            wraps=project._load_urls_progress_listener.loading_resource) as progress_listener_method:
                        # 1. Start server
                        # 2. Fetch page
                        (server_page, _) = await serve_and_fetch_xkcd_home_page(mw)
                        assert 200 == server_page.status
                        
                        # Ensure did not show LoadUrlsProgressDialog
                        assert 0 == progress_listener_method.call_count



from collections import OrderedDict
from crystal.util import cli
from crystal.util.bulkheads import capture_crashes_to_stderr
from crystal.util.wx_bind import bind
from crystal.util.xos import (
    is_linux, is_mac_os, is_windows, windows_major_version,
)
from crystal.util.xthreading import fg_call_later
from enum import Enum
from functools import cached_property
from io import TextIOBase
from threading import Lock
from typing import Dict, List, Optional
import wx
from wx.richtext import RichTextCtrl


def _hexcolor_to_color(hexcolor: str) -> wx.Colour:
    assert hexcolor.startswith('#')
    return wx.Colour(
        int(hexcolor[1:3], 16),
        int(hexcolor[3:5], 16),
        int(hexcolor[5:7], 16)
    )

_CODE_PREFIX = '\033'
_COLOR_CODES = OrderedDict([
    # NOTE: Color hex values determined by matching what
    #       the Terminal app on macOS uses for these color codes
    (cli.TERMINAL_FG_GREEN,  _hexcolor_to_color('#33BD26')),
    (cli.TERMINAL_FG_RED,    _hexcolor_to_color('#C33820')),
    (cli.TERMINAL_FG_YELLOW, _hexcolor_to_color('#AEAD24')),
    (cli.TERMINAL_FG_CYAN,   _hexcolor_to_color('#33BBC7')),
])  # type: Dict[str, wx.Colour]
_RESET_CODE = cli.TERMINAL_RESET


# TODO: Avoid inheriting directly from wx.Frame because doing so
#       implicitly exports a bunch of API from wx.Frame as public
#       that is difficult to support.
class LogDrawer(wx.Frame):
    """
    A drawer that stays attached to the bottom of its parent window,
    even as the parent is repositioned or the parent/drawer is resized.
    
    This drawer contains a textual log that can be written to through the
    file-like object returned by `writer`. The writer understands
    ANSI escape codes generated by the crystal.util.cli package for
    colorizing text.
    
    See also:
    * Positioning and Sizing a Drawer
      https://developer.apple.com/library/archive/documentation/Cocoa/Conceptual/Drawers/Concepts/DrawerSizing.html
    * NSDrawer
      https://developer.apple.com/documentation/appkit/nsdrawer
    """
    
    _INITIAL_LEADING_PLUS_TRAILING_OFFSET = 100  # pixels
    
    # === Init ===
    
    def __init__(self, parent: wx.TopLevelWindow) -> None:
        content_children = self._content_children_of(parent)
        if len(content_children) != 1:
            raise ValueError(
                'LogDrawer can only be attached to a parent with exactly one child. '
                'Suggest moving existing children inside a single wx.Panel container. '
                f'Children: {content_children}')
        
        drawer_style = wx.CLIP_CHILDREN|wx.FRAME_NO_TASKBAR
        if _WindowPairingStrategy.get() == _WindowPairingStrategy.FLOAT_ON_PARENT:
            drawer_style |= wx.FRAME_FLOAT_ON_PARENT
        if wx_resize_border_is_invisible():
            drawer_style |= wx.RESIZE_BORDER
        super().__init__(parent=parent, style=drawer_style)
        
        self._ignore_next_activate = False
        self._leading_plus_trailing_offset = self._INITIAL_LEADING_PLUS_TRAILING_OFFSET
        self._last_height = None  # type: Optional[int]
        self._height_before_closed = None  # type: Optional[int]
        self._parent_will_be_maximized_after_next_resize = False
        self._parent_maximized_size = None
        self._parent_content_container = None  # non-None iff self maximized
        self._was_scrolled_to_bottom = True
        self._is_unmaximizing = False
        
        # Create splitter container that will show a visible sash at the bottom
        # of the drawer, to hint to the user that the drawer is designed
        # to be resized from the bottom edge
        self._splitter = wx.SplitterWindow(
            self,
            style=wx.SP_3D|wx.SP_NO_XP_THEME|wx.SP_LIVE_UPDATE)
        
        # Create upper textarea
        self._textarea = RichTextCtrl(
            self._splitter,
            style=wx.VSCROLL|wx.HSCROLL|wx.NO_BORDER|wx.richtext.RE_READONLY)
        self._textarea.AlwaysShowScrollbars(hflag=False, vflag=True)
        self._textarea.BackgroundColour = wx.Colour(0, 0, 0)  # black
        
        # Create empty bottom
        bottom_pane = wx.Window(self._splitter, style=wx.BORDER_NONE)
        
        # Finish creating splitter container
        self._splitter.SplitHorizontally(
            self._textarea,
            bottom_pane,
            sashPosition=-1  # bottom pane 1 pixel high (the minimum)
        )
        self._splitter.SashGravity = 1.0  # top window grows on resize
        self._splitter.MinimumPaneSize = 1  # prevent unsplit when sash is double-clicked
        
        # Use sizer to delete 1 pixel height used by bottom pane,
        # plus some extra to make the sash look good
        clip_amount = 1  # default
        if is_windows():  # Windows 7, Windows 8, Windows 10
            clip_amount = 4
        elif is_mac_os():  # macOS 10.14
            clip_amount = 3
        self_sizer = wx.BoxSizer(wx.VERTICAL)
        self_sizer.Add(self._splitter, proportion=1, flag=wx.EXPAND|wx.BOTTOM, border=-clip_amount)
        self.SetSizer(self_sizer)
        
        # Reshape self to initial size and position
        desired_width = self.Parent.Size.Width - self._leading_plus_trailing_offset
        self.SetSize(
            x=self._desired_x(desired_width),
            y=self._desired_y,
            width=desired_width,
            height=self._preferred_height,
            sizeFlags=wx.SIZE_USE_EXISTING)
        
        # Set minimum height (when drawer is closed)
        self.SetSizeHints(
            minW=-1,
            maxW=-1,
            minH=self._minimum_height,
            maxH=-1)
        
        # Bind event handlers
        bind(self.Parent, wx.EVT_MOVE, self._on_parent_reshaped)
        bind(self.Parent, wx.EVT_SIZE, self._on_parent_reshaped)
        bind(self.Parent, wx.EVT_MAXIMIZE, self._on_parent_will_maximize)
        if hasattr(wx, 'EVT_FULLSCREEN'):  # wxPython >=4.2.1
            bind(self.Parent, wx.EVT_FULLSCREEN, self._on_parent_will_fullscreen_or_unfullscreen)
        bind(self._textarea, wx.EVT_SCROLLWIN, self._on_scroll)
        
        # Initialize "resized recently" timer
        self._resized_recently_timer = wx.Timer(self)
        bind(self, wx.EVT_TIMER, self._on_resized_recently)
        self._resized_recently_timer.Start(100)  # ms
        
        # Setup drawer's initial maximized state
        if True:
            parent_already_maximized = self.Parent.IsMaximized() or (
                hasattr(self.Parent, 'IsFullScreen') and self.Parent.IsFullScreen())
            
            # Grow parent height to encompass drawer's initial height
            if not parent_already_maximized:
                self.Parent.SetSize(
                    x=wx.DefaultCoord,
                    y=wx.DefaultCoord,
                    width=wx.DefaultCoord,
                    height=self.Parent.Size.Height + self.Size.Height,
                    sizeFlags=wx.SIZE_USE_EXISTING)
            
            # Maximize self initially
            self._on_parent_did_maximize(force_maximize_self=True)
            self.Parent.SendSizeEvent()
    
    def close(self) -> None:
        self._resized_recently_timer.Stop()
        self.Close()
    
    # === Properties ===
    
    @cached_property
    def writer(self) -> TextIOBase:
        """
        Returns a file-like object representing this drawer's content that can be written to.
        
        Writing to the file will write to this drawer's content and will also
        write to sys.stdout.
        """
        return _LogDrawerWriter(self)
    
    # === Shape ===
    
    @property
    def is_open(self) -> bool:
        if self._parent_content_container is None:  # self unmaximized
            return self.Size.Height > self._minimum_height
        else:  # self maximized
            return self._parent_content_container.Window2.Size.Height > self._minimum_content_height
    
    def _desired_x(self, desired_width: int) -> int:
        return int(self.Parent.Position.x + (self.Parent.Size.Width - desired_width)/2)
    
    @property
    def _desired_y(self) -> int:
        return self.Parent.Position.y + self.Parent.Size.Height + self._desired_y_adjustment
    
    @cached_property
    def _desired_y_adjustment(self) -> int:
        if is_windows():
            major_version = windows_major_version()
            if major_version is None:  # unknown Windows
                pass
            elif major_version in [7, 8]:  # Windows 7, Windows 8
                return 0
            elif major_version >= 10:  # Windows 10+
                return -7
        # Default
        return 0
    
    @property
    def _preferred_height(self) -> int:
        PREFERRED_ROW_COUNT = 15
        return self._line_height * PREFERRED_ROW_COUNT
    
    @cached_property
    def _line_height(self) -> int:
        dc = wx.MemoryDC()
        dc.SetFont(self._textarea.Font)
        (_, h) = dc.GetTextExtent('m')
        return h
    
    @cached_property
    def _minimum_height(self) -> int:
        extra_height = 3  # default
        if is_windows():  # Windows *
            extra_height = 3  # Windows 7, Windows 8, Windows 10
        elif is_mac_os():  # macOS 10.14
            extra_height = 2
        result = self._splitter.SashSize + extra_height
        return result
    
    @property
    def _minimum_content_height(self) -> int:
        # NOTE: Must be >= 1 to prevent unsplit
        return 1
    
    # === Actions ===
    
    def _toggle_open(self) -> None:
        """Toggles whether this drawer is open or closed."""
        if self.is_open:
            # Close the drawer
            assert self._parent_content_container is not None  # self maximized
            self._height_before_closed = self._parent_content_container.Window2.Size.Height
            height_after_closed = self._minimum_content_height
            self._parent_content_container.SashPosition += \
                self._height_before_closed - height_after_closed
        else:
            # Open the drawer
            assert self._parent_content_container is not None  # self maximized
            height_after_closed = self._minimum_content_height
            self._parent_content_container.SashPosition -= \
                self._height_before_closed - height_after_closed
            self._height_before_closed = None
    
    def _maximize(self) -> None:
        assert self._parent_content_container is None  # self unmaximized
        
        (parent_content,) = self._content_children_of(self.Parent)
        
        was_closed = not self.is_open  # capture
        
        # Install wx.SplitterWindow in parent
        if True:
            self._parent_content_container = wx.SplitterWindow(self.Parent, style=wx.SP_3D|wx.SP_NO_XP_THEME|wx.SP_LIVE_UPDATE)
            assert self._parent_content_container is not None
            # Prevent unsplit if sash dragged all the way to top or bottom
            self._parent_content_container.MinimumPaneSize = self._minimum_content_height
            self._parent_content_container.SashGravity = 1.0
            parent_content.Reparent(self._parent_content_container)
            self._parent_content_container.Initialize(parent_content)
            self._parent_content_container.Fit()
            
            bind(self._parent_content_container, wx.EVT_SPLITTER_DCLICK, self._on_splitter_double_click)
            bind(self._parent_content_container, wx.EVT_SPLITTER_SASH_POS_CHANGING, self._on_resize_when_maximized)
        
        # Move log from drawer to parent
        if True:
            log_drawer_content = self._splitter.Window1
            log_drawer_content_height = log_drawer_content.Size.Height  # capture
            log_drawer_content.Reparent(self._parent_content_container)
            
            self._parent_content_container.SplitHorizontally(
                self._parent_content_container.Window1,
                log_drawer_content,
                # Target height
                -log_drawer_content_height)
            
            # Adjust drawer content height to match exact target height
            target_height = log_drawer_content_height
            actual_height = log_drawer_content.Size.Height
            self._parent_content_container.SashPosition -= target_height - actual_height
            
            self.Hide()
        
        if was_closed and self.is_open:
            old_height_before_closed = self._height_before_closed  # save
            self._toggle_open()
            assert not self.is_open
            self._height_before_closed = old_height_before_closed  # restore
    
    @property
    def _is_maximized(self) -> bool:
        return self._parent_content_container is not None
    
    @staticmethod
    def _content_children_of(parent: wx.Window) -> list[wx.Window]:
        return [c for c in parent.Children if not isinstance(c, wx.TopLevelWindow)]
    
    # === Events ===
    
    def _on_resize_when_maximized(self, event: wx.SplitterEvent) -> None:
        # Maintain scroll to bottom if applicable
        if self._was_scrolled_to_bottom:
            self._scroll_to_bottom()
        
        # Continue processing event in the normal fashion
        event.Skip()
    
    def _on_resized_recently(self, event: wx.TimerEvent) -> None:
        mouse_state = wx.GetMouseState()
        if not mouse_state.LeftIsDown():
            if self.is_open:
                self._height_before_closed = None
            
            # Stop waiting for mouse up
            self._resized_recently_timer.Stop()
    
    def _on_parent_reshaped(self, event: wx.MoveEvent | wx.SizeEvent) -> None:
        if isinstance(event, wx.SizeEvent):
            if self._parent_will_be_maximized_after_next_resize:
                self._parent_will_be_maximized_after_next_resize = False
                if self._parent_maximized_size is None:
                    self._on_parent_did_maximize()
                else:
                    self._parent_maximized_size = self.Parent.Size
            elif self._parent_maximized_size is not None and self._parent_maximized_size != event.Size:
                self._on_parent_unmaximized()
        
        # Continue processing event in the normal fashion
        event.Skip()
    
    def _on_splitter_double_click(self, event: wx.SplitterEvent) -> None:
        self._toggle_open()
    
    def _on_parent_will_maximize(self, event: wx.MaximizeEvent | None=None) -> None:
        self._parent_will_be_maximized_after_next_resize = True
        
        if event is not None:
            # Continue processing event in the normal fashion
            event.Skip()
    
    def _on_parent_did_maximize(self, *, force_maximize_self: bool=False) -> None:
        self._parent_maximized_size = self.Parent.Size
        if force_maximize_self:
            self._maximize()
    
    def _on_parent_unmaximized(self) -> None:
        self._parent_maximized_size = None
    
    def _on_parent_will_fullscreen_or_unfullscreen(self, event) -> None:
        if event.IsFullScreen():  # fullscreen
            self._parent_will_be_maximized_after_next_resize = True
            self._on_parent_did_maximize()
        else:  # un-fullscreen
            self._on_parent_unmaximized()
        
        # Continue processing event in the normal fashion
        event.Skip()
    
    def _on_scroll(self, event: wx.ScrollEvent) -> None:
        if event.Orientation == wx.VERTICAL:
            self._was_scrolled_to_bottom = self._scrolled_to_bottom
        
        # Continue processing event in the normal fashion
        event.Skip()
    
    @property
    def _scrolled_to_bottom(self) -> bool:
        textarea = self._textarea  # cache
        textarea_was_scrolled_to_bottom = (
            textarea.GetScrollPos(wx.VERTICAL) +
            textarea.GetScrollThumb(wx.VERTICAL) +
            self._line_height
        ) >= textarea.GetScrollRange(wx.VERTICAL)  # capture
        return textarea_was_scrolled_to_bottom
    
    def _scroll_to_bottom(self) -> None:
        textarea = self._textarea  # cache
        textarea.ShowPosition(textarea.LastPosition)


class _LogDrawerWriter(TextIOBase):
    def __init__(self, drawer: LogDrawer) -> None:
        self._drawer = drawer
        self._print_buffer_lock = Lock()
        self._print_buffer = []  # type: List[str]
    
    def write(self, text: str) -> int:
        # Send text to sys.stdout
        print(text, end='')
        
        # Queue text to be printed to drawer
        with self._print_buffer_lock:
            self._print_buffer.append(text)
        
        # Start printing queued texts to drawer, but don't wait until finished
        # 
        # NOTE: Normally would print crashes to LogDrawer, but this IS the
        #       print function for LogDrawer, so fallback to stderr instead.
        @capture_crashes_to_stderr
        def fg_task() -> None:
            with self._print_buffer_lock:
                last_print_buffer = self._print_buffer
                if len(last_print_buffer) == 0:
                    return
                self._print_buffer = []
            
            textarea = self._drawer._textarea  # cache
            textarea_was_scrolled_to_bottom = (
                self._drawer._was_scrolled_to_bottom
            )  # capture
            
            for text in last_print_buffer:
                # Append text to text area
                if True:
                    # Try parse coloring codes around text
                    color: wx.Colour | None = None  # default
                    plain_text: str = text  # default
                    if _CODE_PREFIX in text:
                        if text.endswith(_RESET_CODE):
                            for (code, color) in _COLOR_CODES.items():
                                if text.startswith(code):
                                    break
                            else:
                                # Failed to parse codes
                                color = None
                            if color is not None:
                                plain_text = text[len(code):-len(_RESET_CODE)]
                            else:
                                # Failed to parse codes
                                pass
                        else:
                            # Failed to parse codes
                            pass
                    else:
                        # Text contains no codes
                        pass
                    
                    # Append colored text
                    textarea.SetInsertionPointEnd()
                    if color is not None:
                        textarea.BeginTextColour(color)
                    textarea.WriteText(plain_text)
                    if color is not None:
                        textarea.EndTextColour()
                        # HACK: EndTextColour doesn't seem to actually work
                        #       unless I immediately write some text afterward.
                        # TODO: Eliminate these inserted '\u200b' values
                        #       from any text copied (or cut) from the text area.
                        textarea.WriteText('\u200b')  # zero-width space
            
            if textarea_was_scrolled_to_bottom:
                self._drawer._scroll_to_bottom()
        fg_call_later(fg_task)
        
        # Report successful print
        return len(text)


class _WindowPairingStrategy(Enum):
    """
    Strategy used to keep a drawer window floating directly above 
    (or directly below) its parent window at all times.
    """
    
    # Keep the drawer floating directly above its parent window at all times,
    # using the wx.FRAME_FLOAT_ON_PARENT style bit on the drawer.
    # 
    # Doesn't work properly on macOS: When activating a window
    # from a different app, that window appears below the drawer,
    # rather than that window appearing above both the drawer and
    # the drawer's parent window.
    FLOAT_ON_PARENT = 1
    
    # Keep the drawer floating either directly above or below its parent window at all times,
    # via creative use of the wx.STAY_ON_TOP style bit and wx.Window.Raise().
    FLOAT_AND_RAISE_ON_ACTIVATE = 2
    
    @staticmethod
    def get() -> '_WindowPairingStrategy':
        if is_mac_os():  # macOS 10.14
            return _WindowPairingStrategy.FLOAT_AND_RAISE_ON_ACTIVATE
        elif is_windows():  # Windows 7
            return _WindowPairingStrategy.FLOAT_ON_PARENT
        else:
            return _WindowPairingStrategy.FLOAT_ON_PARENT


def wx_resize_border_is_invisible() -> bool:
    """
    Returns whether the wx.RESIZE_BORDER style applied to a wx.Frame causes
    a resizable border to appear that is visible, as opposed to one that is invisible.
    """
    if is_mac_os():  # macOS 10.14+
        return True
    elif is_windows():  # Windows *
        major_version = windows_major_version()
        if major_version is None:  # unknown Windows
            pass
        elif major_version in [7, 8]:  # Windows 7, Windows 8
            # Windows 7 and 8 add a visible border on all edges
            return False
        elif major_version >= 10:  # Windows 10+
            # Windows 10 adds a visible border on the top edge only
            return False
        else:  # unknown Windows
            pass
        # Unknown Windows
        return False
    elif is_linux():  # Ubuntu 22.04+
        return True
    
    # Default, assuming a newer OS
    return True



import threading
from tqdm.std import TqdmDefaultWriteLock  # type: ignore[attr-defined]
import traceback
from typing import assert_never, Literal


def patch_tqdm_to_debug_deadlocks(on_deadlock: Literal['raise', 'keep_trying']='raise') -> None:
    """
    Patch tqdm to detect potential deadlocks when accessing its internal tqdm._lock.
    
    When a deadlock is detected a warning is printed and one of the following
    actions is taken, depending on the value of `on_deadlock`:
    * 'raise' - Raise an exception
    * 'keep_trying' - Keep trying to acquire the lock anyway.
    """
    TqdmDefaultWriteLock.th_lock_holders = []  # List[Tuple[threading.Thread, traceback.StackSummary]]
    
    def acquire(self, *a, **k):
        k_no_timeout = dict(k)
        if 'timeout' not in k:
            k['timeout'] = 5.0
        for lock in self.locks:
            ok = lock.acquire(*a, **k)
            if not ok:
                try:
                    lock_holder = TqdmDefaultWriteLock.th_lock_holders[-1]
                except IndexError:
                    lock_holder = ('<none>', ['<none>'])
                
                print(
                    f'*** TqdmDefaultWriteLock: Failed to acquire lock in '
                    f'{k["timeout"]} seconds. '
                    f'Lock holder is {lock_holder[0]} at:\n'
                    f'{"".join(lock_holder[1])}')
                print(f'*** New acquirer is:\n{"".join(traceback.format_stack())}')
                if on_deadlock == 'raise':
                    raise AssertionError('Failed to acquire TqdmDefaultWriteLock')
                elif on_deadlock == 'keep_trying':
                    print('*** Continuing to try acquiring lock anyway:')
                    lock.acquire(*a, **k_no_timeout)
                else:
                    assert_never(on_deadlock)
        
        TqdmDefaultWriteLock.th_lock_holders.append((
            threading.current_thread(),
            traceback.format_stack()
        ))
    TqdmDefaultWriteLock.acquire = acquire
    
    super_release = TqdmDefaultWriteLock.release
    def release(self, *args, **kwargs):
        TqdmDefaultWriteLock.th_lock_holders.pop()
        return super_release(self, *args, **kwargs)
    TqdmDefaultWriteLock.release = release



import io
import os.path
import traceback

_CRYSTAL_PACKAGE_PARENT_DIRPATH = os.path.abspath(
    os.path.dirname(os.path.dirname(os.path.dirname(__file__))))

# TODO: Locate the stdlib with a more-reliable method,
#       such as `sysconfig.get_paths().get('stdlib')`.
#       
#       The `os.__file__` method below does not work when Python is running
#       inside a frozen Windows .exe file.
_PYTHON_STDLIB_PARENT_DIRPATH = (
    os.path.abspath(os.path.dirname(os.__file__))
    if hasattr(os, '__file__')
    else None
)


def format_exception_for_user(exc: BaseException) -> str:
    # Try to extract traceback saved by _extract_bulkhead_traceback()
    full_traceback = getattr(exc, '__full_traceback__', Ellipsis)
    if full_traceback is Ellipsis:
        full_traceback = traceback.extract_tb(exc.__traceback__)
    
    exception_str = io.StringIO()
    print(f'{type(exc).__name__}: {str(exc)}', file=exception_str)
    for fs in reversed(full_traceback):
        if '# cr-traceback: ignore' in (fs.line or ''):
            # Don't print frames marked as uninteresting
            continue
        if fs.filename.startswith(_CRYSTAL_PACKAGE_PARENT_DIRPATH):
            short_filepath = os.path.relpath(fs.filename, start=_CRYSTAL_PACKAGE_PARENT_DIRPATH)
        elif (_PYTHON_STDLIB_PARENT_DIRPATH is not None and 
                fs.filename.startswith(_PYTHON_STDLIB_PARENT_DIRPATH)):
            short_filepath = os.path.relpath(fs.filename, start=_PYTHON_STDLIB_PARENT_DIRPATH)
        else:
            short_filepath = fs.filename
        print(f'  at {short_filepath}:{fs.lineno} in {fs.name}', file=exception_str)
    
    return exception_str.getvalue()



# TODO: Consider extracting functions shared between dialogs to own module
from collections.abc import Callable
from crystal.browser.entitytree import ResourceGroupNode, RootResourceNode
from crystal.browser.new_root_url import (
    fields_hide_hint_when_focused, NewRootUrlDialog,
)
from crystal.model import Project, ResourceGroup, ResourceGroupSource
from crystal.progress import CancelLoadUrls
from crystal.util.unicode_labels import decorate_label
from crystal.util.wx_bind import bind
from crystal.util.wx_dialog import (
    CreateButtonSizer, position_dialog_initially, ShowModal,
)
from crystal.util.wx_static_box_sizer import wrap_static_box_sizer_child
from crystal.util.xos import is_linux, is_mac_os, is_windows
from crystal.util.xthreading import fg_affinity
import wx

_WINDOW_INNER_PADDING = 10
_FORM_LABEL_INPUT_SPACING = 5
_FORM_ROW_SPACING = 10
_SCROLL_RATE = 20
_ABOVE_OPTIONS_PADDING = 10

_OPTIONS_SHOWN_LABEL = 'Basic Options'
_OPTIONS_NOT_SHOWN_LABEL = 'Advanced Options'


class NewGroupDialog:
    _INITIAL_URL_PATTERN_WIDTH = NewRootUrlDialog._INITIAL_URL_WIDTH
    _MAX_VISIBLE_PREVIEW_URLS = 100
    _SHOW_SEQUENTIAL_OPTION = False  # hide feature until it is finished
    
    # === Init ===
    
    def __init__(self,
            parent: wx.Window,
            on_finish: Callable[[str, str, ResourceGroupSource, bool, bool], None],
            project: Project,
            saving_source_would_create_cycle_func: Callable[[ResourceGroupSource], bool],
            initial_url_pattern: str='',
            initial_source: ResourceGroupSource=None,
            initial_name: str='',
            initial_do_not_download: bool=False,
            is_edit: bool=False,
            ) -> None:
        """
        Arguments:
        * parent -- parent wx.Window that this dialog is attached to.
        * on_finish -- called when OK pressed on dialog. Is a callable(name, url_pattern, source).
        * project -- the project.
        * initial_url_pattern -- overrides the initial URL pattern displayed.
        * initial_source -- overrides the initial source displayed.
        * initial_name -- overrides the initial name displayed.
        
        Raises:
        * CancelLoadUrls
        """
        self._project = project
        self._on_finish = on_finish
        self._saving_source_would_create_cycle_func = saving_source_would_create_cycle_func
        self._is_edit = is_edit
        
        # Show progress dialog in advance if will need to load all project URLs
        try:
            project.load_urls()
        except CancelLoadUrls:
            raise
        
        dialog = self.dialog = wx.Dialog(
            parent,
            title=('New Group' if not is_edit else 'Edit Group'),
            name='cr-new-group-dialog',
            style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER)
        dialog_sizer = wx.BoxSizer(wx.VERTICAL)
        dialog.SetSizer(dialog_sizer)
        bind(dialog, wx.EVT_BUTTON, self._on_button)
        bind(dialog, wx.EVT_CLOSE, self._on_close)
        
        content_sizer = wx.BoxSizer(wx.VERTICAL)
        content_sizer.Add(
            self._create_fields(dialog, initial_url_pattern, initial_source, initial_name, is_edit),
            flag=wx.EXPAND)
        (
            preview_box,
            preview_box_flags,
            preview_box_border,
        ) = self._create_preview_box(dialog)
        content_sizer.Add(
            preview_box,
            proportion=1,
            flag=wx.EXPAND|preview_box_flags,
            border=preview_box_border)
        
        if not is_edit:
            new_options_sizer = self._create_new_options(dialog)
            content_sizer.Add(
                new_options_sizer,
                flag=wx.EXPAND|wx.TOP,
                border=_ABOVE_OPTIONS_PADDING)
        
        self._advanced_options_sizer = self._create_advanced_options(dialog, initial_do_not_download)
        content_sizer.Add(
            self._advanced_options_sizer,
            flag=wx.EXPAND|wx.TOP,
            border=_ABOVE_OPTIONS_PADDING)
        
        dialog_sizer.Add(content_sizer, proportion=1, flag=wx.EXPAND|wx.ALL,
            border=_WINDOW_INNER_PADDING)
        dialog_sizer.Add(self._create_buttons(dialog, is_edit), flag=wx.EXPAND|wx.BOTTOM,
            border=_WINDOW_INNER_PADDING)
        
        if not fields_hide_hint_when_focused():
            # Initialize focus
            if not is_edit:
                self.pattern_field.SetFocus()
            else:
                self.name_field.SetFocus()
        self._update_preview_urls()
        
        position_dialog_initially(dialog)
        dialog.Fit()
        self._on_options_toggle()  # collapse options initially
        dialog.Show(True)
        
        dialog.MinSize = dialog.Size
        # TODO: Clamp height to fixed value, but still allow
        #       dialog to change height when options shown/hidden
        dialog.MaxSize = wx.Size(wx.DefaultCoord, wx.DefaultCoord)
    
    def _create_fields(self,
            parent: wx.Window,
            initial_url_pattern: str,
            initial_source: ResourceGroupSource,
            initial_name: str,
            is_edit: bool,
            ) -> wx.Sizer:
        fields_sizer = wx.FlexGridSizer(cols=2,
            vgap=_FORM_ROW_SPACING, hgap=_FORM_LABEL_INPUT_SPACING)
        fields_sizer.AddGrowableCol(1)
        
        pattern_field_sizer = wx.BoxSizer(wx.VERTICAL)
        self.pattern_field = wx.TextCtrl(
            parent, value=initial_url_pattern,
            size=(self._INITIAL_URL_PATTERN_WIDTH, wx.DefaultCoord),
            name='cr-new-group-dialog__pattern-field')
        bind(self.pattern_field, wx.EVT_TEXT, self._on_pattern_field_changed)
        self.pattern_field.Hint = 'https://example.com/post/*'
        self.pattern_field.SetSelection(-1, -1)  # select all upon focus
        self.pattern_field.Enabled = not is_edit
        pattern_field_sizer.Add(self.pattern_field, flag=wx.EXPAND)
        pattern_field_sizer.Add(wx.StaticText(parent, label='# = numbers, @ = letters, * = anything but /, ** = anything'), flag=wx.EXPAND)
        
        pattern_label = wx.StaticText(parent, label='URL Pattern:', style=wx.ALIGN_RIGHT|wx.ALIGN_TOP)
        pattern_label.Font = pattern_label.Font.Bold()  # mark as required
        fields_sizer.Add(pattern_label, flag=wx.EXPAND)
        fields_sizer.Add(pattern_field_sizer, flag=wx.EXPAND)
        
        fields_sizer.Add(wx.StaticText(parent, label='Source:', style=wx.ALIGN_RIGHT), flag=wx.EXPAND)
        self.source_choice_box = wx.Choice(
            parent,
            name='cr-new-group-dialog__source-field')
        self.source_choice_box.Append('none', None)
        for rr in self._project.root_resources:
            self.source_choice_box.Append(
                decorate_label(
                    RootResourceNode.ICON,
                    RootResourceNode.calculate_title_of(rr),
                    RootResourceNode.ICON_TRUNCATION_FIX),
                rr)
        for rg in self._project.resource_groups:
            self.source_choice_box.Append(
                decorate_label(
                    ResourceGroupNode.ICON,
                    ResourceGroupNode.calculate_title_of(rg),
                    ResourceGroupNode.ICON_TRUNCATION_FIX),
                rg)
        self.source_choice_box.SetSelection(0)
        if initial_source is not None:
            for i in range(self.source_choice_box.GetCount()):
                cur_source = self.source_choice_box.GetClientData(i)
                if cur_source == initial_source:
                    self.source_choice_box.SetSelection(i)
                    break
        fields_sizer.Add(self.source_choice_box, flag=wx.EXPAND)
        
        fields_sizer.Add(wx.StaticText(parent, label='Name:', style=wx.ALIGN_RIGHT), flag=wx.EXPAND)
        self.name_field = wx.TextCtrl(
            parent, value=initial_name,
            name='cr-new-group-dialog__name-field')
        self.name_field.Hint = 'Post'
        self.name_field.SetSelection(-1, -1)  # select all upon focus
        fields_sizer.Add(self.name_field, flag=wx.EXPAND)
        
        return fields_sizer
    
    def _create_preview_box(self, parent: wx.Window) -> tuple[wx.Window | wx.Sizer, int, int]:
        # NOTE: Don't use wx.CollapsiblePane on wxGTK/Linux because
        #       it doesn't resize its parent window properly on
        #       expand and unexpand events
        preview_box_collapsible = not is_linux()
        preview_box: wx.Window | wx.Sizer
        preview_box_root: wx.Window
        preview_box_root_sizer: wx.BoxSizer
        preview_box_flags: int
        preview_box_border: int
        if preview_box_collapsible:
            preview_box = wx.CollapsiblePane(
                parent, label='Preview Members',
                name='cr-new-group-dialog__preview-members')
            preview_box.Expand()
            preview_box_root = preview_box.GetPane()
            preview_box_root_sizer = wx.BoxSizer(wx.VERTICAL)
            preview_box_root.SetSizer(preview_box_root_sizer)
            preview_box_flags = 0
            preview_box_border = 0
        else:
            preview_box_root_sizer = wx.StaticBoxSizer(
                wx.VERTICAL, parent, label='Preview Members')
            preview_box = preview_box_root_sizer
            preview_box_root = preview_box_root_sizer.GetStaticBox()
            preview_box_flags = wx.TOP
            preview_box_border = 10
        
        preview_box_root_sizer.Add(
            wrap_static_box_sizer_child(
                self._create_preview_box_content(preview_box_root)),
            proportion=1,
            flag=wx.EXPAND)
        
        return (
            preview_box,
            preview_box_flags,
            preview_box_border,
        )
    
    def _create_preview_box_content(self, parent: wx.Window) -> wx.Sizer:
        content_sizer = wx.BoxSizer(wx.VERTICAL)
        
        self.url_list = wx.ListBox(
            parent, style=wx.LB_ALWAYS_SB, size=(-1,150),
            name='cr-new-group-dialog__preview-members__list')
        
        content_sizer.Add(wx.StaticText(parent, label='Known matching URLs:'), flag=wx.EXPAND)
        content_sizer.Add(self.url_list, proportion=1, flag=wx.EXPAND)
        
        return content_sizer
    
    def _create_new_options(self, parent: wx.Window) -> wx.StaticBoxSizer:
        options_sizer = wx.StaticBoxSizer(wx.VERTICAL, parent, label='New Group Options')
        options_sizer.Add(
            wrap_static_box_sizer_child(
                self._create_new_options_content(
                    options_sizer.GetStaticBox())),
            flag=wx.EXPAND)
        return options_sizer
    
    def _create_new_options_content(self, parent: wx.Window) -> wx.Sizer:
        options_sizer = wx.BoxSizer(wx.VERTICAL)
        
        self._download_immediately_checkbox = wx.CheckBox(parent,
            label='Download Group Immediately',
            name='cr-new-group-dialog__download-immediately-checkbox')
        self._download_immediately_checkbox.Value = False
        options_sizer.Add(self._download_immediately_checkbox,
            flag=wx.BOTTOM,
            border=_FORM_LABEL_INPUT_SPACING)
        
        return options_sizer
    
    def _create_advanced_options(self, parent: wx.Window, *args, **kwargs) -> wx.StaticBoxSizer:
        options_sizer = wx.StaticBoxSizer(wx.VERTICAL, parent, label='Advanced Options')
        options_sizer.Add(
            wrap_static_box_sizer_child(
                self._create_advanced_options_content(
                    options_sizer.GetStaticBox(),
                    *args, **kwargs)),
            flag=wx.EXPAND)
        return options_sizer
    
    def _create_advanced_options_content(self,
            parent: wx.Window,
            initial_do_not_download: bool,
            ) -> wx.Sizer:
        options_sizer = wx.BoxSizer(wx.VERTICAL)
        
        self.do_not_download_checkbox = wx.CheckBox(
            parent, label='Do not download members when embedded',
            name='cr-new-group-dialog__do-not-download-checkbox')
        self.do_not_download_checkbox.Value = initial_do_not_download
        options_sizer.Add(self.do_not_download_checkbox,
            flag=wx.BOTTOM,
            border=_FORM_LABEL_INPUT_SPACING)
        
        if self._SHOW_SEQUENTIAL_OPTION:
            options_sizer.AddSpacer(_FORM_ROW_SPACING)
            
            sequential_row = wx.BoxSizer(wx.HORIZONTAL)
            if True:
                self.sequential_checkbox = wx.CheckBox(
                    parent, label='Sequential members',
                    name='cr-new-group-dialog__sequential-checkbox')
                minimum_choice_values = ['0', '1', '2']
                # TODO: Allow the user to specify any integer value, via a wx.ComboBox
                self.minimum_choice_box = wx.Choice(
                    parent,
                    choices=minimum_choice_values)
                self.minimum_choice_box.Selection = minimum_choice_values.index('2')
                
                sequential_row.Add(self.sequential_checkbox, flag=wx.CENTER)
                sequential_row.Add(wx.StaticText(parent, label='–'), flag=wx.CENTER)
                sequential_row.AddSpacer(_FORM_LABEL_INPUT_SPACING)
                sequential_row.Add(wx.StaticText(parent, label='Minimum:'), flag=wx.CENTER)
                sequential_row.AddSpacer(_FORM_LABEL_INPUT_SPACING)
                sequential_row.Add(self.minimum_choice_box, flag=wx.CENTER)
                
            options_sizer.Add(sequential_row)
            
            wildcard_row_scroller = wx.ScrolledWindow(parent, style=wx.HSCROLL|wx.ALWAYS_SHOW_SB)
            if True:
                wildcard_row = wx.BoxSizer(wx.HORIZONTAL)
                if True:
                    url_pattern_fragments = [
                        'https://xkcd.com/',
                        '/',
                        #'/longlonglonglonglonglonglonglonglonglonglonglonglonglonglonglonglonglongmoremoremore',
                    ]
                    url_pattern_wildcards = ['#']
                    assert len(url_pattern_fragments) == len(url_pattern_wildcards) + 1
                    
                    wildcard_labels = [
                        wx.StaticText(wildcard_row_scroller, label=upf)
                        for upf in url_pattern_fragments
                    ]
                    for wl in wildcard_labels:
                        wl.SetForegroundColour(
                            wx.Colour(
                                255 * 5//8,
                                255 * 5//8,
                                255 * 5//8,
                            )  # light gray
                        )
                    
                    wildcard_buttons = [
                        wx.RadioButton(
                            wildcard_row_scroller,
                            label=upw,
                            style=(wx.RB_GROUP if i == 0 else 0))
                        for (i, upw) in enumerate(url_pattern_wildcards)
                    ]
                    wildcard_buttons[0].Value = True  # select first wildcard by default
                    
                    for (wl, wb) in zip(wildcard_labels, wildcard_buttons):
                        wildcard_row.Add(wl, flag=wx.CENTER)
                        wildcard_row.AddSpacer(_FORM_LABEL_INPUT_SPACING)
                        wildcard_row.Add(wb, flag=wx.CENTER)
                        wildcard_row.AddSpacer(_FORM_LABEL_INPUT_SPACING)
                    wildcard_row.Add(wildcard_labels[-1], flag=wx.CENTER)
                
                # NOTE: MUST SetScrollRate to something for scrollbars to work 
                wildcard_row_scroller.SetScrollRate(_SCROLL_RATE, 0)
                
                wildcard_row_scroller.SetSizer(wildcard_row)
                virtual_size = wildcard_row.ComputeFittingWindowSize(wildcard_row_scroller)
                wildcard_row_scroller.SetMinSize(wx.Size(
                    # Prevent scroller's min width from inheriting its virtual min width,
                    # so that the parent can define the width of the scroller
                    0,
                    # Force scroller's min height to match its virtual min height
                    virtual_size.Height
                ))
            
            if is_mac_os():
                checkbox_text_x = 20
            elif is_windows():
                checkbox_text_x = 16
            elif is_linux():
                checkbox_text_x = 24
            else:
                raise AssertionError()
            options_sizer.Add(
                wildcard_row_scroller,
                flag=wx.LEFT|wx.EXPAND,
                # Indent the wildcard row such that its left edge aligns with
                # the "Sequential members" checkbox label
                border=checkbox_text_x)
        
        return options_sizer
    
    def _create_buttons(self, parent: wx.Window, is_edit: bool) -> wx.Sizer:
        ok_button_id = (wx.ID_NEW if not is_edit else wx.ID_SAVE)
        
        button_sizer = wx.BoxSizer(wx.HORIZONTAL)
        
        self._options_button = wx.Button(parent, wx.ID_MORE, _OPTIONS_SHOWN_LABEL)
        button_sizer.Add(
            self._options_button,
            flag=wx.CENTER|wx.LEFT,
            border=_WINDOW_INNER_PADDING)
        
        button_sizer.AddStretchSpacer()
        button_sizer.Add(CreateButtonSizer(parent, ok_button_id, wx.ID_CANCEL), flag=wx.CENTER)
        return button_sizer
    
    # === Operations ===
    
    def _update_preview_urls(self) -> None:
        url_pattern = self.pattern_field.GetValue().strip()
        url_pattern_re = ResourceGroup.create_re_for_url_pattern(url_pattern)
        literal_prefix = ResourceGroup.literal_prefix_for_url_pattern(url_pattern)
        
        url_pattern_is_literal = (len(literal_prefix) == len(url_pattern))
        if url_pattern_is_literal:
            member = self._project.get_resource(literal_prefix)
            if member is None:
                (matching_urls, approx_match_count) = ([], 0)
            else:
                (matching_urls, approx_match_count) = ([member.url], 1)
        else:
            (matching_urls, approx_match_count) = self._project.urls_matching_pattern(
                url_pattern_re, literal_prefix, limit=self._MAX_VISIBLE_PREVIEW_URLS)
        
        self.url_list.Clear()
        if len(matching_urls) > 0:  # avoid warning on Mac
            more_count = approx_match_count - len(matching_urls)
            more_items = (
                [f'... about {more_count:n} more']
                if more_count != 0
                else []
            )
            self.url_list.InsertItems(matching_urls + more_items, 0)
    
    # === Events ===
    
    def _on_pattern_field_changed(self, event) -> None:
        self._update_preview_urls()
    
    @fg_affinity
    def _on_button(self, event: wx.CommandEvent) -> None:
        btn_id = event.GetEventObject().GetId()
        if btn_id in (wx.ID_NEW, wx.ID_SAVE):
            self._on_ok()
        elif btn_id == wx.ID_CANCEL:
            self._on_cancel()
        elif btn_id == wx.ID_MORE:
            self._on_options_toggle()
    
    @fg_affinity
    def _on_close(self, event: wx.CloseEvent) -> None:
        self._on_cancel()
    
    @fg_affinity
    def _on_ok(self) -> None:
        name = self.name_field.Value
        url_pattern = self.pattern_field.Value.strip()
        if len(url_pattern) == 0:
            dialog = wx.MessageDialog(
                self.dialog,
                message='Please enter a URL pattern.',
                caption='Empty URL Pattern',
                style=wx.OK,
            )
            dialog.Name = 'cr-empty-url-pattern'
            position_dialog_initially(dialog)
            choice = ShowModal(dialog)
            assert wx.ID_OK == choice
            return
        source = self.source_choice_box.GetClientData(
            self.source_choice_box.GetSelection())
        if self._saving_source_would_create_cycle_func(source):
            dialog = wx.MessageDialog(
                self.dialog,
                message='Cannot use that Source because it has this group as a Source.',
                caption='Source Cycle Created',
                style=wx.OK,
            )
            dialog.Name = 'cr-source-cycle-created'
            position_dialog_initially(dialog)
            choice = ShowModal(dialog)
            assert wx.ID_OK == choice
            return
        do_not_download = self.do_not_download_checkbox.Value
        if self._is_edit:
            download_immediately = False
        else:
            download_immediately = self._download_immediately_checkbox.Value
        self._on_finish(
            name, url_pattern, source, do_not_download,
            download_immediately)
        
        self.dialog.Destroy()
    
    @fg_affinity
    def _on_cancel(self) -> None:
        self.dialog.Destroy()
    
    @fg_affinity
    def _on_options_toggle(self) -> None:
        options = self._advanced_options_sizer.GetStaticBox()
        if options.Shown:
            # Hide
            self._options_button.Label = _OPTIONS_NOT_SHOWN_LABEL
            
            options_height = options.Size.Height + _ABOVE_OPTIONS_PADDING
            options.Shown = False
            self.dialog.SetSize(
                x=wx.DefaultCoord,
                y=wx.DefaultCoord,
                width=wx.DefaultCoord,
                height=self.dialog.Size.Height - options_height,
                sizeFlags=wx.SIZE_USE_EXISTING)
        else:
            # Show
            self._options_button.Label = _OPTIONS_SHOWN_LABEL
            
            options.Shown = True
            options_height = options.Size.Height + _ABOVE_OPTIONS_PADDING
            self.dialog.SetSize(
                x=wx.DefaultCoord,
                y=wx.DefaultCoord,
                width=wx.DefaultCoord,
                height=self.dialog.Size.Height + options_height,
                sizeFlags=wx.SIZE_USE_EXISTING)



from collections.abc import Callable, Iterator
from contextlib import contextmanager
from crystal.util import cli
from functools import wraps
import sys
import traceback
from typing import Concatenate, overload, Protocol, TypeVar
from typing_extensions import ParamSpec

_S = TypeVar('_S')
_B = TypeVar('_B', bound='Bulkhead')
_P = ParamSpec('_P')
_R = TypeVar('_R')
_RT = TypeVar('_RT')
_RF = TypeVar('_RF')


# ------------------------------------------------------------------------------
# Bulkheads

CrashReason = BaseException  # with .__traceback__ set to a TracebackType


class Bulkhead(Protocol):  # abstract
    """
    A sink for unhandled exceptions (i.e. crashes).
    """
    crash_reason: CrashReason | None


class BulkheadCell(Bulkhead):
    """
    A concrete Bulkhead which stores any crash that occurs,
    but takes no special action to report such crashes.
    """
    crash_reason: CrashReason | None
    
    def __init__(self, value: CrashReason | None=None) -> None:
        self.crash_reason = value


@overload
def capture_crashes_to_self(
        bulkhead_method: Callable[Concatenate[_B, _P], _RT]
        ) -> Callable[Concatenate[_B, _P], _RT | None]:
    ...

@overload
def capture_crashes_to_self(
        *, return_if_crashed: _RF
        ) -> Callable[[Callable[Concatenate[_B, _P], _RT]], Callable[Concatenate[_B, _P], _RT | _RF]]:
    ...

@overload
def capture_crashes_to_self(
        ) -> Callable[[Callable[Concatenate[_B, _P], _RT]], Callable[Concatenate[_B, _P], _RT | None]]:
    ...

def capture_crashes_to_self(
        bulkhead_method: Callable[Concatenate[_B, _P], _RT] | None=None,
        *, return_if_crashed=None  # _RF
        ):
    """
    A Bulkhead method that captures any raised exceptions to itself,
    as the "crash reason" of the bulkhead.
    
    If the bulkhead was already crashed (with a non-None "crash reason") then
    this method will immediately abort, returning `return_if_crashed`.
    
    Examples:
        class MyBulkhead(Bulkhead):
            @capture_crashes_to_self
            def foo_did_bar(self) -> None:
                ...
            
            @capture_crashes_to_self(return_if_crashed=Ellipsis)
            def calculate_foo(self) -> Result:
                ...
    """
    def decorate(
            bulkhead_method: Callable[Concatenate[_B, _P], _RT]
            ) -> Callable[Concatenate[_B, _P], _RT | _RF]:
        @wraps(bulkhead_method)
        @_mark_bulkhead_call
        def bulkhead_call(self: _B, *args: _P.args, **kwargs: _P.kwargs) -> _RT | _RF:
            if self.crash_reason is not None:
                # Bulkhead has already crashed. Abort.
                return return_if_crashed
            try:
                return bulkhead_method(self, *args, **kwargs)  # cr-traceback: ignore
            except BaseException as e:
                e.__full_traceback__ = _extract_bulkhead_traceback(e)  # type: ignore[attr-defined]
                
                # Print traceback to assist in debugging in the terminal,
                # including ancestor callers of bulkhead_call
                _print_bulkhead_exception(e)
                
                # Crash the bulkhead. Abort.
                self.crash_reason = e
                return return_if_crashed
        return bulkhead_call
    if bulkhead_method is None:
        return decorate
    else:
        return decorate(bulkhead_method)


@overload
def capture_crashes_to_bulkhead_arg(
        method: Callable[Concatenate[_S, _B, _P], _RT]
        ) -> Callable[Concatenate[_S, _B, _P], _RT | None]:
    ...

@overload
def capture_crashes_to_bulkhead_arg(
        *, return_if_crashed: _RF
        ) -> Callable[[Callable[Concatenate[_S, _B, _P], _RT]], Callable[Concatenate[_S, _B, _P], _RT | _RF]]:
    ...

@overload
def capture_crashes_to_bulkhead_arg(
        ) -> Callable[[Callable[Concatenate[_S, _B, _P], _RT]], Callable[Concatenate[_S, _B, _P], _RT | None]]:
    ...

def capture_crashes_to_bulkhead_arg(
        method: Callable[Concatenate[_S, _B, _P], _RT] | None=None,
        *, return_if_crashed=None  # _RF
        ):
    """
    A method that captures any raised exceptions to its first Bulkhead argument,
    as the "crash reason" of the bulkhead.
    
    If the bulkhead was already crashed (with a non-None "crash reason") then
    this method will immediately abort, returning `return_if_crashed`.
    
    Examples:
        class MyClass:
            @capture_crashes_to_bulkhead_arg
            def other_foo_did_bar(self, other: Bulkhead) -> None:
                ...
            
            @capture_crashes_to_bulkhead_arg(return_if_crashed=Ellipsis)
            def calculate_baz(self, other: Bulkhead) -> Result:
                ...
    """
    def decorate(
            method: Callable[Concatenate[_S, _B, _P], _RT]
            ) -> Callable[Concatenate[_S, _B, _P], _RT | _RF]:
        @wraps(method)
        @_mark_bulkhead_call
        def bulkhead_call(self: _S, bulkhead: _B, *args: _P.args, **kwargs: _P.kwargs) -> _RT | _RF:
            if bulkhead.crash_reason is not None:
                # Bulkhead has already crashed. Abort.
                return return_if_crashed
            try:
                return method(self, bulkhead, *args, **kwargs)  # cr-traceback: ignore
            except BaseException as e:
                e.__full_traceback__ = _extract_bulkhead_traceback(e)  # type: ignore[attr-defined]
                
                # Print traceback to assist in debugging in the terminal,
                # including ancestor callers of bulkhead_call
                _print_bulkhead_exception(e)
                
                # Crash the bulkhead. Abort.
                bulkhead.crash_reason = e
                return return_if_crashed
        return bulkhead_call
    if method is None:
        return decorate
    else:
        return decorate(method)


@overload
def capture_crashes_to(
        bulkhead: Bulkhead
        ) -> Callable[[Callable[_P, _RT]], Callable[_P, _RT | None]]:
    ...

@overload
def capture_crashes_to(
        bulkhead: Bulkhead,
        return_if_crashed: _RF
        ) -> Callable[[Callable[_P, _RT]], Callable[_P, _RT | _RF]]:
    ...

def capture_crashes_to(
        bulkhead: Bulkhead,
        return_if_crashed=None  # _RF
        ) -> Callable[[Callable[_P, _RT]], Callable[_P, _RT | _RF]]:
    """
    A method that captures any raised exceptions to the specified Bulkhead,
    as the "crash reason" of the bulkhead.
    
    If the bulkhead was already crashed (with a non-None "crash reason") then
    this method will immediately abort, returning `return_if_crashed`.
    
    Examples:
        @capture_crashes_to(bulkhead)
        def foo_did_bar() -> None:
            ...
        
        @capture_crashes_to(bulkhead, return_if_crashed=Ellipsis)
        def calculate_foo() -> Result:
            ...
    """
    def decorate(func: Callable[_P, _RT]) -> Callable[_P, _RT | _RF]:
        @wraps(func)
        @_mark_bulkhead_call
        def bulkhead_call(*args: _P.args, **kwargs: _P.kwargs) -> _RT | _RF:
            if bulkhead.crash_reason is not None:
                # Bulkhead has already crashed. Abort.
                return return_if_crashed
            try:
                return func(*args, **kwargs)  # cr-traceback: ignore
            except BaseException as e:
                e.__full_traceback__ = _extract_bulkhead_traceback(e)  # type: ignore[attr-defined]
                
                # Print traceback to assist in debugging in the terminal,
                # including ancestor callers of bulkhead_call
                _print_bulkhead_exception(e)
                
                # Crash the bulkhead. Abort.
                bulkhead.crash_reason = e
                return return_if_crashed
        return bulkhead_call
    return decorate


@contextmanager
def crashes_captured_to(bulkhead: Bulkhead, *, enter_if_crashed: bool=False) -> Iterator[None]:
    """
    Context that captures any raised exceptions to the specified Bulkhead,
    as the "crash reason" of the bulkhead.
    
    If the bulkhead was already crashed (with a non-None "crash reason")
    when entering the context, the contents of the context will be skipped,
    unless enter_if_crashed=True.
    
    Example:
        with crashes_captured_to(bulkhead, enter_if_crashed=True):
            ...
    """
    if not enter_if_crashed:
        # NOTE: It's probably not actually possible to implement the
        #       enter_if_crashed=False case because the context manager
        #       protocol provides no reasonable way to skip the interior
        #       context entirely.
        raise NotImplementedError()
    try:
        yield  # cr-traceback: ignore
    except BaseException as e:
        e.__full_traceback__ = _extract_bulkhead_traceback(  # type: ignore[attr-defined]
            e, fix_tb=lambda here_tb, exc_tb: here_tb[:-3] + exc_tb[1:])
        
        # Print traceback to assist in debugging in the terminal
        _print_bulkhead_exception(e)
        
        # Crash the bulkhead. Abort.
        bulkhead.crash_reason = e
        return


@overload
def capture_crashes_to_stderr(
        func: Callable[_P, _RT]
        ) -> Callable[_P, _RT | None]:
    ...

@overload
def capture_crashes_to_stderr(
        *, return_if_crashed: _RF
        ) -> Callable[[Callable[_P, _RT]], Callable[_P, _RT | _RF]]:
    ...

@overload
def capture_crashes_to_stderr(
        ) -> Callable[[Callable[_P, _RT]], Callable[_P, _RT | None]]:
    ...

def capture_crashes_to_stderr(
        func: Callable[_P, _RT] | None=None,
        *, return_if_crashed=None  # _RF
        ):
    """
    A method that captures any raised exceptions, and prints them to stderr.
    
    Examples:
        @capture_crashes_to_stderr
        def foo(self) -> None:
            ...
        
        @capture_crashes_to_stderr(return_if_crashed=Ellipsis)
        def calculate_foo(self) -> Result:
            ...
    """
    def decorate(func: Callable[_P, _RT]) -> Callable[_P, _RT | _RF]:
        @wraps(func)
        @_mark_bulkhead_call
        def bulkhead_call(*args: _P.args, **kwargs: _P.kwargs) -> _RT | _RF:
            try:
                return func(*args, **kwargs)  # cr-traceback: ignore
            except BaseException as e:
                e.__full_traceback__ = _extract_bulkhead_traceback(e)  # type: ignore[attr-defined]
                
                # Print traceback to assist in debugging in the terminal,
                # including ancestor callers of bulkhead_call
                _print_bulkhead_exception(e, is_error=True)
                
                # Abort.
                return return_if_crashed
        return bulkhead_call
    if func is None:
        return decorate
    else:
        return decorate(func)


def does_not_capture_crashes(func: Callable[_P, _R]) -> Callable[_P, _R]:
    """
    Explicitly marks functions that intentionally do not capture crashes,
    allowing them to bubble up to their caller.
    """
    return func


def _mark_bulkhead_call(bulkhead_call: Callable[_P, _R]) -> Callable[_P, _R]:
    bulkhead_call._captures_crashes = True  # type: ignore[attr-defined]
    return bulkhead_call


_ExtractedTraceback = list[traceback.FrameSummary]
_FixTbFunc = Callable[[_ExtractedTraceback, _ExtractedTraceback], _ExtractedTraceback]

def _extract_bulkhead_traceback(
        e: BaseException,
        *, fix_tb: _FixTbFunc | None=None,
        extra_stacklevel: int=0
        ) -> _ExtractedTraceback | None:
    if e.__traceback__ is None:
        return None
    here_tb = traceback.extract_stack(sys._getframe(1 + extra_stacklevel))
    exc_tb = traceback.extract_tb(e.__traceback__)
    if fix_tb is None:
        fix_tb = lambda here_tb, exc_tb: here_tb[:-1] + exc_tb
    full_tb_summary = fix_tb(here_tb, exc_tb)  # type: _ExtractedTraceback
    return full_tb_summary


def _print_bulkhead_exception(e: BaseException, *, is_error: bool=False, fix_tb: _FixTbFunc | None=None) -> None:
    # Print traceback to assist in debugging in the terminal,
    # including ancestor callers of bulkhead_call
    err_file = sys.stderr
    print(
        cli.TERMINAL_FG_RED if is_error else cli.TERMINAL_FG_YELLOW,
        end='', file=err_file)
    full_tb_summary = getattr(e, '__full_traceback__', Ellipsis)
    if full_tb_summary is Ellipsis:
        full_tb_summary = _extract_bulkhead_traceback(
            e, fix_tb=fix_tb, extra_stacklevel=1)
    if full_tb_summary is not None:
        print('Exception in bulkhead:', file=err_file)
        print('Traceback (most recent call last):', file=err_file)
        for x in traceback.format_list(full_tb_summary):
            print(x, end='', file=err_file)
    for x in traceback.format_exception_only(type(e), e):
        print(x, end='', file=err_file)
    print(cli.TERMINAL_RESET, end='', file=err_file)
    err_file.flush()


def run_bulkhead_call(
        bulkhead_call: Callable[_P, _R],
        /, *args: _P.args,
        **kwargs: _P.kwargs
        ) -> '_R':
    """
    Calls a function marked as @capture_crashes_to*,
    which does not reraise exceptions from its interior.
    
    Raises AssertionError if the specified function is not actually
    marked with @capture_crashes_to*.
    """
    ensure_is_bulkhead_call(bulkhead_call)
    return bulkhead_call(*args, **kwargs)  # cr-traceback: ignore


def ensure_is_bulkhead_call(callable: Callable) -> None:
    """
    Raises AssertionError if the specified function is not actually
    marked with @capture_crashes_to*.
    """
    if not is_bulkhead_call(callable):
        raise AssertionError(f'Expected callable {callable!r} to be decorated with @capture_crashes_to*')


def is_bulkhead_call(callable: Callable) -> bool:
    """
    Returns whether the specified function is marked with @capture_crashes_to*.
    """
    return getattr(callable, '_captures_crashes', False) == True


# ------------------------------------------------------------------------------